#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 BloxOne Threat Defence API Wrapper Class and Helpers

 Author: Chris Marrison

 Date Last Updated: 20211203

 Todo:

 Copyright (c) 2020-2021 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.3.3'
__author__ = 'Chris Marrison'
__author_email__ = 'chris@infoblox.com'

import bloxone
import logging
import datetime
import json


# b1td class
class b1td(bloxone.b1):
    '''
    BloxOne ThreatDefence API Wrapper
    Covers TIDE and Dossier
    '''

    # Generic Methods
    def get(self, objpath, **params):
        '''
        Generic get object wrapper for TIDE data objects

        Parameters:
            objpath (str):  Swagger object path
            action (str):   Optional object action
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tide_url + objpath
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response


    def post(self, objpath, body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.ddi_url + objpath

        # Make API Call
        response = self._apipost(url, body)

        return response


   # *** Helper Methods ***

    def threat_classes(self, **params):
        '''
        Get list of threat classes

        Parameters:

        Returns:
            response object: Requests response object
        '''

        # Local Variables
        objpath = '/data/threat_classes'

        # Build url
        url = self.tide_url + objpath

        # Make API Call
        response = self._apiget(url)

        return response


    def threat_properties(self, threatclass="", **params):
        '''
        Get list of threat properties

        Parameters:
            threatclass (str): Threat Class

        Returns:
            response object: Requests response object
        '''

        # Local Variables
        objpath = '/data/properties'

        # Build url
        url = self.tide_url + objpath
        if threatclass:
            url = url + '?class=' + threatclass

        # Make API Call
        response = self._apiget(url)

        return response

    def threat_counts(self):
        '''
        Query Infoblox TIDE for active threat counts

        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.tide_url + '/data/threat/counts'

        # Make API Call
        response = self._apiget(url)

        return response


    def historical_threat_counts(self):
        '''
        Query Infoblox TIDE for historical threat counts

        Returns:
            response object: Requests response object
        '''
        # Build URL
        url = self.tide_url + '/data/threat/counts/historical'

        # Make API Call
        response = self._apiget(url)

        return response


    def default_ttl(self):
        '''
        '''
        url = self.tide_url + '/data/default/ttl'
        response = self._apiget(url)

        return response


    def querytide(self, datatype, query, **params):
        '''
        Query Infoblox TIDE for all avaialble threat data
        related to query.

        Parameters:
            datatype (str): "host", "ip" or "url"
            query (str): query data

        Returns:
            response object: Requests response object
        '''
        objpath = '/data/threats/'

        # Build URL
        url = self.tide_url + objpath
        url = url + datatype + '?' + datatype + '=' + query
        url = self._add_params(url, first_param=False, **params)

        # Make API Call
        response = self._apiget(url)

        return response


    def querytideactive(self, datatype, query, **params):
        '''
        Query Infoblox TIDE for "active" threat data
        i.e. threat data that has not expired at time of call

        Parameters:
            datatype (str): "host", "ip" or "url"
            query (str): query data

        Returns:
            response object: Requests response object
        '''
        objpath = '/data/threats/'
        now = datetime.datetime.now()

        # Build URL
        url = self.tide_url + objpath
        url = url + datatype + '?' + datatype + '=' + query
        url = url + '&expiration=' + now.strftime('%Y-%m-%dT%H:%M:%SZ')
        url = self._add_params(url, first_param=False, **params)

        # Make API Call
        response = self._apiget(url)

        return response


    def querytidestate(self, datatype, query, **params):
        '''
        Query Infoblox TIDE State Tables for specific query

        Parameters:
            datatype (str): "host", "ip" or "url"
            query (str): query data

        Returns:
            response object: Requests response object
        '''
        objpath = '/data/threats/state/'

        # Build URL
        url = self.tide_url + objpath
        url = url + datatype + '?' + datatype + '=' + query
        url = self._add_params(url, first_param=False, **params)

        # Make API Call
        response = self._apiget(url)

        return response


    def tideactivefeed(self, 
                       datatype, 
                       profile="", 
                       threatclass="",
                       threatproperty="",
                       **params ):
        '''
        Bulk "active" threat intel download from Infoblox TIDE state tables
        for specified datatype.

        Parameters:
            datatype (str): "host", "ip" or "url"
            profile (str, optional): Data provider
            threatclass (str, optional): tide data class
            threatproperty (str, optional): tide data property

        Returns:
            response object: Requests response object
        '''
        objpath = '/data/threats/state/'

        # Build URL
        url = self.tide_url + objpath
        url = url + datatype 
        if profile or threatclass or threatproperty:
            url = url + '?'
            if profile:
                url = url + '&profile=' + profile
            if threatclass:
                url = url + '&class=' + threatclass
            if threatproperty:
                url = url + '&property=' + threatproperty
            url = self._add_params(url, first_param=False, **params) 
        else:
            url = self._add_params(url, **params) 

        # Make API Call
        response = self._apiget(url)

        return response


    def tidedatafeed(self, 
                    datatype, 
                    profile="", 
                    threatclass="",
                    threatproperty="",
                    **params):
        '''
        Bulk threat intel download from Infoblox TIDE 
        for specified datatype. Please use wisely.

        Parameters:
            datatype (str): "host", "ip" or "url"
            profile (str, optional): Data provider
            threatclass (str, optional): tide data class
            threatproperty (str, optional): tide data property

        Returns:
            response object: Requests response object
        '''
        objpath = '/data/threats/'

        # Build URL
        url = self.tide_url + objpath
        url = url + datatype 
        if profile or threatclass or threatproperty:
            url = url + '?'
            if profile:
                url = url + '&profile=' + profile
            if threatclass:
                url = url + '&class=' + threatclass
            if threatproperty:
                url = url + '&property=' + threatproperty
            url = self._add_params(url, first_param=False, **params) 
        else:
            url = self._add_params(url, **params) 

        # Make API Call
        response = self._apiget(url)

        return response

    # ** Dossier Methods **

    def dossier_sources(self):
        '''
        Get Sources for Dossier

        Returns:
            response object: Requests response object
        '''
        url = self.dossier_url + '/sources'
        response = self._apiget(url)

        return response


    def dossier_target_sources(self, type='host'):
        '''
        Get supported target types for Dossier 

        Parameters:
            type (str): target type

        Returns:
            response object: Request response object
        '''
        url = self.dossier_url + '/sources/target/' + type
        response = self._apiget(url)

        return response


    def dossier_target_types(self):
        '''
        Get supported target types for Dossier 

        Returns:
            response object: Request response object
        '''
        url = self.dossier_url + '/targets'
        response = self._apiget(url)

        return response


    def dossierquery(self, query, type="host", sources="all", wait=True):
        '''
        Simple Dossier Query
        
        Parameters:
            query (str or list): single query or list of same type
            type (str): "host", "ip" or "url"
            sources (str): set of sources or "all"

        Returns:
            response object: Requests response object
        '''

        if isinstance(wait, bool) and wait:
            wait = 'true'
        else:
            wait = 'false'
        url = self.dossier_url + '/jobs?wait=' + wait
        # Create body
        if sources == "all":
            response = self.dossier_target_sources(type=type)
            if response.status_code in self.return_codes_ok:
                sources = []
                for source in response.json().keys():
                    if response.json()[source]:
                        sources.append(source)
                logging.debug("Sources retrieved: {}".format(sources))
            else:
                sources = ['atp', 'activity', 'dns', 'geo', 'pdns', 'ptr', 
                           'rwhois', 'whopis', 'inforank', 'rpz_feeds', 
                           'ssl_cert', 'infoblox_web_cat', 'custom_lists',
                           'whitelist']
                logging.debug("Failed to retrieve sources, using " 
                              + "limited list {}".format(sources))
        else:
            if not isinstance(sources, list):
                source_list = []
                source_list.append(sources) 
                sources = source_list

        # Check for group of queries
        if isinstance(query, list): 
            body = {"target": {"group": {"type": type,
                    "sources": sources,
                    "targets": query  } } }
        else:
            body = {"target": {"one": {"type": type,
                    "sources": sources, "target": query  } } }

        body = json.dumps(body)
        logging.debug(f'Body: {body}')

        response = self._apipost(url, body)

        return response


    # *** threat enrichment 

    def threat_actor(self, name):
        '''
        Get Threat Actor details

        Parameters:
            name(str): Name of Threat Actor
        
        Returns:
            response object: Requests response object
        '''

        url = self.threat_enrichment_url + '/threat_actor/lookup?name=' + name
        response = self._apiget(url)

        return response


    def expand_mitre_vector(self, mitre):
        '''
        Expand MITRE Vector details

        Parameters:
            mitre(str): MITRE Vector
        
        Returns:
            response object: Requests response object
        '''

        url = self.threat_enrichment_url + '/mitre/lookup'
        body = '"' + mitre + '"'
        logging.debug("URL: {}, Body: {}".format(url,body))
        response = self._apipost(url, body)

        return response

    def get_countries(self):
        '''
        Get Countries and Country Code Data

        Parameters:
            None
        
        Returns:
            response object: Requests response object
        '''
        url = self.tide_url + '/data/set/countries'
        logging.debug("URL: {}".format(url))

        return self._apiget(url)


    def get_country_isocode(self, country=""):
        '''
        Get ISO Code for specified country

        Parameters:
            country (str): Name of Country
        
        Returns:
            isocode (str): ISO Code of Country or None if no match 
        '''
        isocode = ''

        response = self.get_countries()
        if response.status_code in self.return_codes_ok:
            country_codes = response.json()['country']
            country_record = next((c for c in country_codes 
                                if c['name'].casefold() == country.casefold()), 
                                {} )
            isocode = country_record.get('iso_code')

        else:
            logging.error(f'Unable to retrieve country list')
            isocode = None
        
        return isocode
            

    def get_country_ips(self, country='', **params):
        '''
        Get IPs for specified countries or complete dataset

        Parameters:
            country: Country or Country Code to retrieve
        
        Returns:
            response object: Requests response object
        
        Raises:
            CountryISOCodeNotFound
        '''
        iso_code = ''
        url = self.tide_url + '/data/set/countryip'
        
        if country:
            if len(country) == 2:
                # Assume country isocode
                iso_code = country
            else:
                # Assume country name
                iso_code = self.get_country_isocode(country=country)
            if iso_code:
                url = self._add_params(url, first_param=True, country=iso_code)
            else:
                raise self.CountryISOCodeNotFound(f'No match for country: {country}')

        return self._apiget(url)





from .bloxone import *
from .b1anycast import *
from .b1authn import *
from .b1bootstrap import *
from .b1ddi import *
from .b1diagnostics import *
from .b1cdc import *
from .b1notifications import *
from .b1infrastructure import *
from .b1insights import *
from .b1oph import *
from .b1platform import *
from .b1sw import *
from .b1ntp import *
from .b1td import *
from .b1tdc import *
from .b1tdep import *
from .b1tddfp import *
from .b1tdlad import *
from .b1user import *
from .b1ztp import *
from .utils import *
from .dhcputils import *

__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__version__ = '0.9.0rc3'



#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 Module to provide class hierachy to simplify access to the BloxOne APIs

 Date Last Updated: 20230222

 Todo:

 Copyright (c) 2021-2023 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
import bloxone
import logging
import json

# ** Global Vars **
__version__ = '0.2.5'
__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__doc__ = 'https://python-bloxone.readthedocs.io/en/latest/'
__license__ = 'BSD'


class b1oph(bloxone.b1):
    '''
    Class to simplify access to the BloxOne Platform APIs
    '''

    def __init__(self, cfg_file='config.ini'):
        '''
        Call base __init__ and extend
        '''
        super().__init__(cfg_file)

        self.OPH_COMPOSITE_STATUS = { '0': 'Review Details', 
                                      '1': 'Online',
                                      '2': 'Unknown',
                                      '3': 'Pending',
                                      '4': 'Awaiting approval' }
        
        self.OPH_PLATFORM_STATES = { '0': 'Offline',
                                     '1': 'Online',
                                     '2': 'Error',
                                     '3': 'Waiting (pending)',
                                     '4': 'Unknown' }
        
        self.OPH_HOST_TYPES = { '0': 'Not Available',
                                '1': 'Unknown',
                                '2': 'Unknown',
                                '3': 'BloxOne VM',
                                '4': 'BloxOne Appliance - B105',
                                '5': 'BloxOne Container',
                                '6': 'CNIOS' }
        
        self.OPH_APP_MGT = { '0': 'Inactive',
                             '1': 'Active',
                             '2': 'Error',
                             '3': '' }

        self.OPH_APPS = { '1': { 'AppName': 'DFP', 'StatusSpace': '9' },
                          '2': { 'AppName': 'DNS', 'StatusSpace': '12' },
                          '3': { 'AppName': 'DHCP', 'StatusSpace': '15' },
                          '7': { 'AppName': 'CDC', 'StatusSpace': '24' },
                          '9': { 'AppName': 'Anycast', 'StatusSpace': '30' },
                          '10': { 'AppName': 'NGC', 'StatusSpace': '34' },
                          '12': { 'AppName': 'MS AD Collector', 'StatusSpace': '40' },
                          '13': { 'AppName': 'Access Authentication', 'StatusSpace': '43' },
                          '14': { 'AppName': 'Edge_Services_FW', 'StatusSpace': '46' },
                          '15': { 'AppName': 'Edge_Services_Router', 'StatusSpace': '49' },
                          '16': { 'AppName': 'Site-to-Site_VPN', 'StatusSpace': '52' },
                          '18': { 'AppName': 'DNS_Assured_Forwarding', 'StatusSpace': '58' },
                          '20': { 'AppName': 'NTP', 'StatusSpace': '64' }
        }

        self.OPH_APP_NAMES = { 'DFP': '1',
                               'DNS': '2',
                               'DHCP': '3',
                               'CDC': '7',
                               'Anycast': '9',
                               'NGC': '10' }

        self.APP_STATUS = { '0': 'inactive', '1': 'active', '2': 'stopped' }

        logging.warning('The b1oph class is being deprecated, please see b1infrastructure class')

        return


    def get(self, objpath, id="", action="", **params):
        '''
        Generic get object wrapper for platform calls

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"

        Returns:
            response object: Requests response object
        '''

        # Build url
        url = self.host_url + objpath
        url = self._use_obj_id(url,id=id)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response

        
    def create(self, objpath, body=""):
        '''
        Generic create object wrapper for platform objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.host_url + objpath
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, objpath, id=""):
        '''
        Generic delete object wrapper for platform objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to delete

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.host_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apidelete(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.host_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response


    def patch(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.host_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipatch(url, body)

        return response


    def get_tags(self, objpath, id=""):
        '''
        Get tags for an object id

        Parameters:
            objpath (str):  Swagger object path

            id (str): id of object

        Returns:
            tags (dict): Dictionary of current tags
                         or empty dict if none
        
        .. todo::
            * make generic, however, this requires the below
            * Lookup dictionary of 'required fields' per object type
        '''
        tags = {}
        response = self.get(objpath, id=id, _fields="tags")
        if response.status_code in self.return_codes_ok:
            tags = json.loads(response.text)
            tags = tags['result']
        else:
            tags = {}
        
        return tags


    # *** Platform API Requests *** 

    def on_prem_hosts(self, **params):
        '''
        Method to retrieve On Prem Hosts
        (undocumented)

        Parameters:
            **params (dict): Generic API parameters

        Returns:
            response object: Requests response object
        '''

        # Call BloxOne API
        response = self.get('/on_prem_hosts', **params)

        # Return response object
        return response 


    def get_id(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id = ""
        filter = key+'=="'+value+'"'
        fields = key + ',id'

        # Make API Call
        response = self.get(objpath, _filter=filter, _fields=fields)

        # Process response
        if response.status_code in self.return_codes_ok:
            obj = response.json()
            # Look for results or result
            if "results" in obj.keys():
                result_text = "results"
            elif "result" in obj.keys():
                result_text = "result"
            else:
                result_text = ""
            
            # Get result if available
            if result_text:
                obj = obj[result_text]
                if obj:
                    id = obj[0]['id']
                    if not include_path and "/" in str(id):
                        id = id.rsplit('/',1)[1]
                else:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


    def get_ophid(self, name=""):
        '''
        Return the ophid of named OPH

        Parameters:
            name (str): display name of OPH
        
        Returns:
            ophid(str): ophid of the specified OPH
        '''
        filter = f'display_name=="{name}"'
        response = self.get('/on_prem_hosts', _filter=filter, _fields="display_name,ophid")
        if response.status_code in self.return_codes_ok:
            if 'result' in response.json().keys():
                ophid = response.json()['result'][0]['ophid']
            else:
                ophid = ''
        else:
            ophid = ''

        return ophid


    def oph_add_tag(self, id="", tagname="", tagvalue=""):
        '''
        Method to add a tag to an existing On Prem Host

        Parameters:
            objpath (str):  Swagger object path
            tagname (str): Name of tag to add
            tagvalue (str): Value to associate

        Returns:
            response object: Requests response object
        '''
        # tags = self.get_tags('/on_prem_hosts', id=id)
        response = self.get('/on_prem_hosts', id=id, _fields="display_name,tags")
        if response.status_code in self.return_codes_ok:
            data = response.json()['result']
        else:
            data = {}
        logging.debug("Existing tags: {}".format(data))
        # Add new tag to data
        if tagname:
            data['tags'].update({tagname: tagvalue})
            logging.debug("New tags: {}".format(data))
        # Update object
        response = self.update('/on_prem_hosts', id=id, body=json.dumps(data))

        return response


    def oph_delete_tag(self, id="", tagname=""):
        '''
        Method to delete a tag from an existing On Prem Host

        Parameters:
            objpath (str):  Swagger object path
            tagname (str): Name of tag to add

        Returns:
            response object: Requests response object
        '''
        # tags = self.get_tags('/on_prem_hosts', id=id)
        response = self.get('/on_prem_hosts', id=id, _fields="display_name,tags")
        if response.status_code in self.return_codes_ok:
            data = response.json()['result']
            logging.debug("Existing tags: {}".format(data))
            # Delete tag from data
            if tagname in data['tags'].keys():
                data['tags'].pop(tagname, True)
                print(json.dumps(data))
                logging.debug("New tags: {}".format(data))
                # Update object
                response = self.update('/on_prem_hosts', id=id, body=json.dumps(data))

        return response


    def oph_status_summary(self, name="", id="", tags=False):
        '''
        Get OPH status information for one or more OPHs

        Parameters:
            name (ste): Display name of OPH
            id (str): id of a specific OPH
            tags (bool): include tags in report
        
        Returns:
            rpt: Dict of translated status elements
        '''
        rpt = None

        # Get OPH Data
        if name:
            name_filter = 'display_name=="' + name + '"'
            response = self.get('/on_prem_hosts', _filter=name_filter)
        elif id:
            response = self.get('/on_prem_hosts', id=id)
        else:
            response = self.get('/on_prem_hosts')
        
        if response.status_code in self.return_codes_ok:
            rpt = self.oph_status_rpt(response.json(), tags=tags)
        else:
            logging.error(f'Response code: {response.status_code}, Data: {response.text}')
            rpt = None
    
        return rpt


    def oph_status(self, oph_data):
        '''
        Translate status info in JSON data for an OPH

        Parameters:
            oph_data (dict): Data for individual OPH
        
        Returns:
            oph_status: Dict of translated status elements
        '''
        oph_status = {}
        oph_comp_state = ''
        oph_plat_state = ''
        oph_plat_msg = ''
        oph_app_mgt = ''
        oph_app_msg = ''

        if 'composite_status' in oph_data.keys():
            oph_comp_state = self.OPH_COMPOSITE_STATUS[oph_data['composite_status'].get('status')]
        else:
            oph_comp_state = 'No state information'
                        
        if 'state' in oph_data.keys():
            if oph_data['state'].get('current_state') in self.OPH_PLATFORM_STATES.keys():
                oph_plat_state = self.OPH_PLATFORM_STATES[oph_data['state'].get('current_state')]
            else:
                oph_plat_state = f"Unknown state: {oph_data['state'].get('current_state')}"

            if 'message' in oph_data['state'].keys():
                oph_plat_state = oph_data['state']['message']

        else:
            oph_plat_state = "Platform status unavailable"
        
        if 'status' in oph_data.keys():
            oph_app_mgt = self.OPH_PLATFORM_STATES[oph_data['status'].get('status')]
            if 'message' in oph_data['status'].keys():
                oph_app_msg = oph_data['status']['message']
        else:
            oph_app_mgt = "Application Management status unavailable"

        oph_status = { 'OPH State': oph_comp_state,
                       'Platform Management': oph_plat_state,
                       'Application Management': oph_app_mgt }
        
        if oph_plat_msg:
            oph_status.update( {'Platform Message': oph_plat_msg })
        if oph_app_msg:
            oph_status.update( {'App Mgt Message': oph_app_msg })
        
        return oph_status
    

    def oph_app_status(self, oph_data):
        '''
        Translate App status info in JSON data for an OPH

        Parameters:
            oph_data (dict): Data for individual OPH
        
        Returns:
            oph_apps: Dict of translated status elements
        '''
        oph_apps = {}
        app_id = ''
        app_name = ''
        ver_name = ''
        status = ''
        app_version = None

        # Check for application status data
        if 'applications' in oph_data.keys():
            for app in oph_data['applications']:
                app_id = app.get('application_type')
                if app_id in self.OPH_APPS.keys():
                    app_name = self.OPH_APPS[app_id]['AppName']
                    ver_name = app_name + '_version'
                else:
                    app_name = 'Unknown, AppID: ' + app_id

                # Check whether app is enabled
                status = self.APP_STATUS[app['composite_status']['status']]
                if app.get('disabled') == '1':
                    status = f'disabled - {status}'

                app_version = app.get('current_version')

                # Add version info if available
                if app_version:
                    oph_apps.update( { app_name: status,
                                        ver_name: app_version } )
                else:
                    oph_apps.update( { app_name: status } )

        else:
            oph_apps = {}
        
        return oph_apps
   

    def oph_status_rpt(self, json_data, tags=False):
        '''
        Build  status report from GET /on_prem_hosts data

        Parameters:
            json_data (json): API JSON data for On Prem Hosts call
            tags (bool): Include tags in response, default False
        
        Returns:
            rpt: Dict of status elements
        '''
        rpt = {}
        on_prem_host = ""
        oph_status = {}
        oph_apps = {}

        if json_data:
            results = json_data.get('result')
            if not isinstance(results, list):
                results = [ results ]
            # Build report 
            for oph in results:
                oph_apps = {}
                on_prem_host = oph.get('display_name')
                if on_prem_host:
                    oph_status = self.oph_status(oph)
                    oph_apps = self.oph_app_status(oph)

                    rpt.update( { on_prem_host: { 'status': oph_status,
                                          'id': oph.get('id'),
                                          'host_type': self.OPH_HOST_TYPES[oph.get('host_type')],
                                          'ip_address': oph.get('ip_address'), 
                                          'nat_ip': oph.get('nat_ip'),
                                          'version': oph.get('current_version'),
                                          'last_seen': oph.get('last_seen'),
                                          'applications': oph_apps }
                                } )
                if tags:
                    rpt[on_prem_host].update({'tags': oph.get('tags')})
        else:
            rpt = None

        return rpt

    
    def oph_uptime(self, name=""):
        '''
        '''
        uptime = None
        ophid = self.get_ophid(name=name)
        url = 'https://csp.infoblox.com/atlas-status-service/v1/getsummary'
        body = '{"objectID":["' + ophid +'"],"objectType":["Onprem Host ID"],"event_key":["health-collector/heartbeat/"]}'
        response = self._apipost(url, body=body)
        if response.status_code in self.return_codes_ok:
            uptime = response.json()['results'][0]['metadata']['hostUptime']
        
        return uptime


    def get_app_state(self, name, app):
        '''
        Get status of application for an OPH

        Parameters:
            name (str): display_name of OPH
            app (str): App Name, e.g. DNS
        
        Returns:
            app_status (str): Status or error msg as text
        '''
        status_data = ''
        app_status = ''

        status_data = self.oph_status_summary(name=name)
        if status_data:
            if 'applications' in status_data[name].keys():
                app_status = status_data[name]['applications'].get(app)
                if not app_status:
                    logging.error(f'App: {app} not found for OPH: {name}')
                    app_status = f'App: {app} not found for OPH: {name}'
            else:
                logging.error(f'No application data for OPH: {name}')
                app_status = f'No application data for OPH: {name}'

        else:
            logging.error(f'OPH: {name} not found')
            app_status = f'OPH: {name} not found'

        return app_status


    def manage_app(self, name="", app="", action="status"):
        '''
        Perform action on named OPH for specified app

        Parameters:
            name (str): display_name of OPH
            app (str): App Name, e.g. DNS
            action (str): action to perform for app
        
        Returns:
            bool
        '''
        result = False
        actions = [ 'status', 'disable', 'enable', 'stop', 'start' ]

        if action in actions:
           if action == "status":
               result = self.get_app_state(name=name, app=app) 
           elif action == "disable":
               result = self.disable_app(name=name, app=app) 
           elif action == "enable":
               result = self.enable_app(name=name, app=app) 
           elif action == "start":
               result = self.app_process_control(name=name, 
                                                 app=app, 
                                                 action="start") 
           elif action == "stop":
               result = self.app_process_control(name=name, 
                                                 app=app, 
                                                 action="stop") 
           else:
               logging.error(f'Action: {action} not implemented')
               result = False
        else:
            logging.error(f'Action: {action} not supported')
            logging.info(f'Supported actions: {actions}')
            result = False

        return result


    def disable_app(self, name="", app=""):
        '''
        Disable specified app on named OPH

        Parameters:
            name (str): display_name of OPH
            app (str): App Name, e.g. DNS
        
        Returns:
            bool
        '''
        status = False
        app_type = ''

        # Check app supported and get Get application_type
        if app in self.OPH_APP_NAMES.keys():
            app_type = self.OPH_APP_NAMES[app]
        elif app in self.OPH_APPS.keys():
            app_type = app

        if app_type:
            # Get id of OPH
            filter = f'display_name=="{name}"'
            response = self.get('/on_prem_hosts', 
                                _filter=filter)
            if response.status_code in self.return_codes_ok:
                oph_data = response.json()['result']
                id = oph_data[0]['id']
                logging.debug(f'On Prem Host id = {id}')

                # Build body
                body = { 'display_name': name, 
                         'applications': [ { 'application_type': app_type,
                                             'disabled': '1', 
                                             'state': { 'desired_state': '0' } 
                                           } ] } 
                
                # Update desired OPH
                response = self.update('/on_prem_hosts',
                                       id=id,
                                       body=json.dumps(body))
                if response.status_code in self.return_codes_ok:
                    logging.debug(f'OPH: {name}, App: {app}, App_type: {app_type}')
                    status = True
                else:
                    logging.error(f'{response.status_code}: {response.text}')

            else:
                logging.error(f'OPH {name} not found.')
                status = False
        else:
            logging.error(f'App {app} not supported.')
            status = False
        
        return status


    def enable_app(self, name="", app=""):
        '''
        Enable specified app on named OPH

        Parameters:
            name (str): display_name of OPH
            app (str): App Name, e.g. DNS
        
        Returns:
            bool
        '''
        status = False
        app_type = ''

        # Check app supported and get Get application_type
        if app in self.OPH_APP_NAMES.keys():
            app_type = self.OPH_APP_NAMES[app]
        elif app in self.OPH_APPS.keys():
            app_type = app

        if app_type:
            # Get id of OPH
            filter = f'display_name=="{name}"'
            response = self.get('/on_prem_hosts', 
                                _filter=filter)
            if response.status_code in self.return_codes_ok:
                oph_data = response.json()['result']
                id = oph_data[0]['id']
                logging.debug(f'On Prem Host id = {id}')

                # Build body
                body = { 'display_name': name, 
                         'applications': [ { 'application_type': app_type,
                                             'disabled': '0', 
                                             'state': { 'desired_state': '1' } 
                                           } ] } 
                
                # Update desired OPH
                response = self.update('/on_prem_hosts',
                                       id=id,
                                       body=json.dumps(body))
                if response.status_code in self.return_codes_ok:
                    logging.debug(f'OPH: {name}, App: {app}, App_type: {app_type}')
                    status = True
                else:
                    logging.error(f'{response.status_code}: {response.text}')

            else:
                logging.error(f'OPH {name} not found.')
                status = False
        else:
            logging.error(f'App {app} not supported.')
            status = False
        
        return status
    

    def app_process_control(self, name="", app="", action=""): 
        '''
        Start or stop an application process

        Parameters:
            name (str): display_name of OPH
            app (str): App Name, e.g. DNS
        
        Returns:
            bool 
        '''
        app_type = ''
        status = False
        actions = { "start": '1', "stop": '0' }

        if action in actions.keys():

            # Check app supported and get Get application_type
            if app in self.OPH_APP_NAMES.keys():
                app_type = self.OPH_APP_NAMES[app]
            elif app in self.OPH_APPS.keys():
                app_type = app

            if app_type:
                # Get id of OPH
                oph_status = self.oph_status_summary(name=name)
                if oph_status:
                    id = oph_status.get('id')
                    logging.debug(f'On Prem Host id = {id}')
                    app_status = oph_status[name]['applications'].get(app)
                    logging.debug(f'App status: {app} app is {app_status}')
                    # Check whether app is disabled
                    if 'disabled' not in app_status:
                        # Build body
                        body = { 'display_name': name, 
                                'applications': [ { 'application_type': app_type,
                                                    'disabled': '0', 
                                                    'state': 
                                                        { 'desired_state': actions[action] } 
                                                } ] } 
                        
                        # Update desired OPH
                        response = self.update('/on_prem_hosts',
                                            id=id,
                                            body=json.dumps(body))
                        if response.status_code in self.return_codes_ok:
                            logging.debug(f'OPH: {name}, App: {app}, App_type: {app_type}')
                            status = True
                        else:
                            logging.error(f'{response.status_code}: {response.text}')
                            status = False
                    else:
                        logging.info(f'App: {app} on OPH: {name} {app_status}')
                        status = False

                else:
                    logging.error(f'OPH {name} not found.')
                    status = False
            else:
                logging.error(f'App {app} not supported.')
                status = False
        else:
            logging.error(f'Action: {action} not supported')
            status = False

        return status



#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:
    Simple utility functions for data type validation, domain handling,
    and data normalisationa specifically with the aim of supporting
    queries to TIDE and Dossier.

 Requirements:
   Python3 with re, ipaddress, requests 

 Author: Chris Marrison

 Date Last Updated: 20210804

 Todo:

 Copyright (c) 2021 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.2.7'
__author__ = 'Chris Marrison'
__author_email__ = 'chris@infoblox.com'

import logging
import ipaddress
import os
import yaml
import binascii
import bloxone
from pprint import pprint

# ** Global Vars **
# DHCP Encoding Utils

class dhcp_encode():
    '''
    Class to assist with Hex Encoding of 
    DHCP Options and sub_options
    '''
    def __init__(self) -> None:
        self.opt_types = [ 'string', 
                           'ipv4_address', 
                           'ipv6_address',
                           'boolean',
                           'int8',
                           'uint8',
                           'int16',
                           'uint16',
                           'int32',
                           'uint32',
                           'fqdn',
                           'binary',
                           'empty' ]

        self.fqdn_re, self.url_re = bloxone.utils.buildregex()
        return


    def validate_ip(self, ip):
        '''
        Validate input data is a valid IP address
        (Supports both IPv4 and IPv6)

        Parameters:
            ip (str): ip address as a string

        Returns:
            bool: Return True for valid and False otherwise

        '''
        try:
            ipaddress.ip_address(ip)
            result = True
        except ValueError:
            result = False
        return result


    # IP encondings
    def ip_to_hex(self, ip):
        '''
        Encode an IPv4 or IPv6 address to hex

        Parameters:
            ip (str): IPv4 or IPv6 address as a string
        
        Returns:
            hex encoding as string
        '''
        if self.validate_ip(ip):
            ip = ipaddress.ip_address(ip)
            hex_value = '{:x}'.format(ip)
        else:
            logging.error(f'{ip} not a valid IP')
            hex_value = ''

        return hex_value

    # Methods for IPv4 and IPv6
    def ipv4_address_to_hex(self, ipv4):
        '''
        Encode an IPv4 address to hex

        Parameters:
            ipv4 (str): IPv4 address as a string
        
        Returns:
            hex encoding as string
        '''
        return self.ip_to_hex(ipv4)
        
    
    def ipv6_address_to_hex(self, ipv6):
        '''
        Encode an IPv6 address to hex

        Parameters:
            ipv6 (str): IPv4 or IPv6 address as a string
        
        Returns:
            hex encoding as string
        '''
        return self.ip_to_hex(ipv6)


    # String/text encoding
    def string_to_hex(self, string):
        '''
        Encode a text string to hex

        Parameters:
            string (str): text string
        
        Returns:
            hex encoding as string
        '''
        s = str(string).encode('utf-8')
        return s.hex()


    # Boolean encoding
    def boolean_to_hex(self, flag):
        ''' 
        Encode boolean value as single hex byte

        Parameters:
            flag (bool/str): True or False as bool or text
        
        Returns:
            hex encoding as string
        '''
        # Handle Bool or str
        if not isinstance(flag, bool):
            if isinstance(flag, str):
                if flag.casefold() == 'true': 
                    flag = True
                else:
                    flag = False
            else:
                flag = False

        # Set hex value     
        if flag:
            hex = '01'
        else:
            hex = '00'
        
        return hex


   # integer encodings 
    def int_to_hex(self, i, size = 8):
        '''
        Encode integer of specified size as signed int in hex

        Parameters:
            i (int): integer value to encode
            size (int): size in bits [8, 16, 32]
        
        Returns:
            hex encoding as string
        '''
        hex_value = ''
        i = int(i)
        i_sizes = [ 8, 16, 32 ]
        if size in i_sizes:
            max_bits = size - 1
            no_bytes = size // 4
            fmt = f'{{:0{no_bytes}x}}'
            if i > 0 and i < (2**max_bits):
                hex_value = fmt.format(int(i))
            elif abs(i) <= (2**max_bits):
                hex_value = fmt.format(int(abs(i) + (2**max_bits))) 
            else:
                raise TypeError(f'{i} is out of range for int{size} type')

        return hex_value 


    def uint_to_hex(self, i, size = 8):
        '''
        Encode integer of specified size as unsigned int in hex
        Uses 2's compliment if supplied with negative number


        Parameters:
            i (int): integer value to encode
            size (int): size in bits [8, 16, 32]
        
        Returns:
            hex encoding as string
        '''
        i = int(i)
        i_sizes = [ 8, 16, 32 ]
        if size in i_sizes:
            max_size = 2**size
            no_octets = size // 4
            fmt = f'{{:0{no_octets}x}}'
            if i < 0 and abs(i) < max_size:
                hex_str = fmt.format(i + (2**size))
            elif i < max_size:
                hex_str = fmt.format(i)
            else:
                raise TypeError(f'{i} is out of range for uint{size} type')

        return hex_str 


    # Methods for intX and uintX
    def int8_to_hex(self, value):
        return self.int_to_hex(value, size=8)


    def uint8_to_hex(self, value):
        return self.uint_to_hex(value, size=8)


    def int16_to_hex(self, value):
        return self.int_to_hex(value, size=16)


    def uint16_to_hex(self, value):
        return self.uint_to_hex(value, size=16)


    def int32_to_hex(self, value):
        return self.int_to_hex(value, size=32)


    def uint32_to_hex(self, value):
        return self.uint_to_hex(value, size=32)


    # FDQN Encoding
    def fqdn_to_hex(self, fqdn):
        '''
        Encode an fdqn in RFC 1035 Section 3.1 formatted hex

        Parameters:
            fqdn (str): hostname to encode

        Returns:
            hex encoding as string
        '''
        hex = ''
        hex_label = ''

        # Validate FQDN
        if bloxone.utils.validate_fqdn(fqdn, self.fqdn_re):
            if fqdn.endswith("."):
                # strip exactly one dot from the right, if present
                fqdn = fqdn[:-1]
            # Encode labels
            for label in fqdn.split("."):
                hex_label = self.string_to_hex(label)    
                hex_len = self.hex_length(hex_label)
                hex += hex_len + hex_label

            # Terminate with null byte
            hex += '00'

        else:
            logging.error(f'{fqdn} is not a valid FQDN')
        
        return hex


    # Binary Encoding
    def binary_to_hex(self, data):
        '''
        Format hex string of binary/hex encoded data

        Parameters:
            data (str): data to format

        Returns:
            hex encoding as string
        '''
        hex_value = ''
        base = 16 

        # Check for binary
        if data[:2] == '0b':
            base = 2
        
        # Force hex encoding without 0x using base
        hex_value = '{:02x}'.format(int(data, base))

        return hex_value

    
    # Empty Encoding
    def empty_to_hex(self, data):
        '''
        Return empyt hex string ''

        Parameters:
            data (str): Data not to encode (should be empty)
        
        Returns:
            Empty String ''
        '''
        if data:
            data = ''
        return data
        

    # Code and Length encoding
    def optcode_to_hex(self, optcode):
        '''
        Encode Option Code in hex (1-octet)

        Parameters:
            optcode (str/int): Option Code 
        
        Returns:
            hex encoding as string
        '''
        hex_opt = '{:02x}'.format(int(optcode))
        return hex_opt


    def hex_length(self, hex_string):
        '''
        Encode Option Length in hex (1-octet)

        Parameters:
            hex_string (str): Octet Encoded Hex String
        
        Returns:
            Number of Hex Octects as hex encoded string
        '''
        hex_len = '{:02x}'.format(int(len(hex_string) / 2))
        return hex_len


    # Encoding Methods
    def encode_data(self, sub_opt, padding=False, pad_bytes=1):
        '''
        Encode the data section of a sub_option definition

        Parameters:
            sub_opt (dict): Dict containing sub option details
                            Must include 'data' and 'type' keys
            padding (bool): Whether extra 'null' termination bytes are req.
            pad_bytes (int): Number of null bytes to append

        Returns:
            Hex encoded data for specified data-type as string
        '''
        hex_data = ''
        if sub_opt['type'] in self.opt_types:
            type_to_hex = eval('self.' + sub_opt['type'] + '_to_hex')
        else:
            logging.error(f'Unsupported Option Type {sub_opt["type"]}')
            logging.info('Unsupported option type, ' +
                         'attempting to process as string')
            type_to_hex = eval('self.string_to_hex')
        
        # Check for array attribute
        if 'array' in sub_opt.keys():
            array = sub_opt['array']
        else:
            logging.debug(f'No array attribute for option: {sub_opt["code"]}')
            array = False

        # Encode data
        if array:
            for item in sub_opt['data'].split(','):
                hex_data += type_to_hex(item)
        else:
            hex_data = type_to_hex(sub_opt['data'])       
        
        if padding:
            hex_data += pad_bytes * '00'

        return hex_data


    def encode_sub_option(self, sub_opt, 
                          data_only=False,
                          padding=False,
                          pad_bytes=1):
        '''
        Encode individual sub option

        Parameters:
            sub_opt (dict): Sub Option Definition, as dict.
            data_only (bool): Encode data portion only if True
                              (Note the sub_opt dict is also checked for 
                              the 'data-only' key)
            padding (bool): Whether extra 'null' termination bytes are req.
            pad_bytes (int): Number of null bytes to append
        
        Returns:
            Encoded suboption as a hex string
        '''
        # Local variables
        hex_value = ''
        hex_opt = ''
        hex_len = ''
        hex_data = ''

        # Check whether 'data-only' is defined in sub_option
        if 'data-only' in sub_opt.keys():
            data_only = sub_opt['data-only']
        # Check whether to encode only the data or whole sub option
        if data_only:
            hex_data = self.encode_data(sub_opt)
            hex_value = hex_data
        else:
            if int(sub_opt['code']) in range(0, 256):
                hex_opt = self.optcode_to_hex(sub_opt['code'])
                hex_data = self.encode_data(sub_opt)
                hex_len = self.hex_length(hex_data)
                hex_value = hex_opt + hex_len + hex_data
            else:
                # Log error (or potentially raise exception or something)
                logging.error(f'Option Code: {sub_opt["code"]} out of range.')
                hex_value = ''
        
        return hex_value


    def encode_dhcp_option(self, 
                           sub_opt_defs=[], 
                           padding=False,
                           pad_bytes=1,
                           encapsulate=False,
                           id=None,
                           prefix='' ):
        '''
        Encode list of DHCP Sub Options to Hex

        Parameters:
            sub_opt_defs (list): List of Sub Option definition dictionaries
            padding (bool): Whether extra 'null' termination bytes are req.
            pad_bytes (int): Number of null bytes to append
            encapsulate (bool): Add id and total length as prefix
            id (int): option code to prepend
            prefix (str): String value to prepend to encoded options
        
        Returns:
            Encoded suboption as a hex string
        '''
        hex_value = ''

        # Encode sub_options
        for opt in sub_opt_defs:
            hex_value += self.encode_sub_option(opt) 
        
        if encapsulate:
            total_len = self.hex_length(hex)
            main_opt = self.optcode_to_hex(id)
            hex_value = main_opt + total_len + hex_value
        
        if prefix:
            hex_value += prefix 

        return hex_value


    def tests(self):
        '''
        Run through encoding methods and output example results
        '''
        test_data = [ { 'code': '1', 'type': 'string',
                        'data': 'AABBDDCCEEDD-aabbccddeeff' },
                      { 'code': '2', 'type': 'ipv4_address',
                        'data': '10.10.10.10' },
                      { 'code': '3', 'type': 'ipv4_address',
                        'data': '10.10.10.10,11.11.11.11', 'array': True },
                      { 'code': '4', 'type': 'boolean', 'data': True },
                      { 'code': '5', 'type': 'int8', 'data': '22' },
                      { 'code': '5', 'type': 'int8', 'data': '-22' },
                      { 'code': '6', 'type': 'uint8', 'data': '22' },
                      { 'code': '7', 'type': 'int16', 'data': '33'},
                      { 'code': '8', 'type': 'int16', 'data': '33'},
                      { 'code': '9', 'type': 'uint16', 'data': '33'}, 
                      { 'code': '10', 'type': 'int32', 'data': '44'}, 
                      { 'code': '11', 'type': 'uint32', 'data': '-44'}, 
                      { 'code': '12', 'type': 'uint32', 'data': '44'}, 
                      { 'code': '13', 'type': 'fqdn',
                        'data': 'www.infoblox.com' },
                      { 'code': '14', 'type': 'binary', 'data': 'DEADBEEF' },
                      { 'code': '15', 'type': 'empty', 'data': ''},
                      { 'code': '16', 'type': 'ipv6_address',
                        'data': '2001:DB8::1' },
                      { 'code': '17', 'type': 'ipv6_address',
                        'data': '2001:DB8::1,2001:DB8::2', 'array': True } ]

        print(f'Encoding types supported: {self.opt_types}')
        print()
        print('Data tests:')
        for data_test in test_data:
            result = self.encode_data(data_test)
            hex_len = self.hex_length(result)
            print(f'Type: {data_test["type"]}: {data_test["data"]}, ' +
                  f'Encoded: {result}, Length(hex): {hex_len}')
        
        print()
        # Padding Test
        test_data = { 'code': '99', 'type': 'string', 'data': 'AABBCCDD' }
        result = self.encode_data(test_data, padding=True)
        print(f'Padding test (1 byte), type string: {test_data["data"]}' +
              f' {result}')
        # Full encode test
        test_data = [ { 'code': '1', 'type': 'string',
                        'data': 'AABBDDCCEEDD-aabbccddeeff' },
                      { 'code': '2', 'type': 'ipv4_address',
                        'data': '10.10.10.10' },
                      { 'code': '3', 'type': 'ipv4_address',
                        'data': '10.10.10.10,11.11.11.11', 'array': True },
                      { 'code': '4', 'type': 'boolean', 'data': True },
                      { 'code': '5', 'type': 'int8', 'data': '22' } ]
        result = self.encode_dhcp_option(test_data)
        print(f'Full encoding of sample: {result}')

        return
                

# *** Class to handle Vendor Option Definitions Dictionary in YAML ***
class DHCP_OPTION_DEFS():
    '''
    Class to load and handle DHCP Option Defs
    '''

    def __init__(self, cfg='vendor_dict.yaml'):
        '''
        Initialise Class Using YAML config

        Parameters:
            cfg (str): Config file to load, default vendor_dict.yaml
        
        Raises:
            FileNotFoundError is yaml file is not found
        '''
        self.config = {}
   
        # Check for yaml file and raise exception if not found
        if os.path.isfile(cfg):
            # Read yaml configuration file
            try:
                self.config = yaml.safe_load(open(cfg, 'r'))
            except yaml.YAMLError as err:
                logging.error(err)
                raise
        else:
            logging.error(f'No such file {cfg}')
            raise FileNotFoundError(f'YAML config file "{cfg}" not found.')

        return


    def version(self):
        '''
        Returns:
        
            str containing config file version or 'Version not defined'
        '''
        if 'version' in self.config.keys():
            version = self.config['version']
        else:
            version = 'Version not defined'

        return version


    def keys(self):
        '''
        Returns:
            list of top level keys
        '''
        return self.config.keys()


    def vendors(self):
        '''
        Returns:
           list of defined vendors
        '''
        return self.config['vendors'].keys()


    def vendor_keys(self, vendor):
        '''
        Returns vendor top level keys

        Parameters:
            vendor (str): Vendor Identifier
        
        Returns:
            list of keys defined for a vendor
        '''
        if self.included(vendor):
             response = self.config['vendors'][vendor].keys()
        else:
            response = None
        
        return response


    def count(self):
        '''
        Get numbner of defined vendors
    
        Returns:
            int
        '''
        return len(self.config['vendors'])  


    def included(self, vendor):
        '''
        Check whether this vendor is configured

        Parameters:
            vendor (str): Vendor Identifier

        Returns bool
        '''
        status = False
        if vendor in self.vendors():
            status = True
        else:
            status = False

        return status


    def vendor_description(self, vendor):
        '''
        Get description of vendor

        Parameters:
            vendor (str): Vendor Identifier
        '''
        desc = None
        if self.included(vendor):
            desc = self.config['vendors'][vendor]['description']
        else:
            desc = None
        
        return desc


    def vendor_prefix(self, vendor):
        '''
        Get the prefix is present as a string

        Parameters:
            vendor (str): Vendor Identifier
        
        Returns:
            string containing defined prefix or '' if none
        '''
        prefix = ''
        if self.included(vendor):
            if 'prefix' in self.vendor_keys(vendor):
                prefix = self.config['vendors'][vendor]['prefix']
        
        return prefix


    def option_def(self, vendor):
        '''
        Returns option definition as dict

        Parameters:
            vendor (str): Vendor Identifier
        
        Returns:
            Dict containing both parent and sub option definitions
        '''
        opt_def = {}
        if self.included(vendor):
            if 'option-def' in self.vendor_keys(vendor):
                opt_def = self.config['vendors'][vendor]['option-def']
            else:
                logging.error(f'No option definition for vendor {vendor}')
        else:
            logging.error(f'Vendor: {vendor} not defined')
        
        return opt_def


    def parent_opt_def(self, vendor):
        '''
        Returns parent-option definition as dict

        Parameters:
            vendor (str): Vendor Identifier
        
        Returns:
            dict containing parent option definition
        '''
        opt_def = {}
        parent_def = {}
        if self.included(vendor):
            opt_def = self.option_def(vendor)
            if 'parent-option' in opt_def.keys():
                parent_def = opt_def['parent-option']
            else:
                logging.error(f'No parent-option for vendor {vendor}')
        else:
            logging.error(f'Vendor: {vendor} not defined')
        
        return parent_def


    def sub_options(self, vendor):
        '''
        Returns list of sub-option definitions

        Parameters:
            vendor (str): Vendor Identifier
        
        Returns:
            list of dict
        '''
        opt_def = {}
        sub_opt_defs = []
        if self.included(vendor):
            opt_def = self.option_def(vendor)
            if 'sub-options' in opt_def.keys():
                sub_opt_defs = opt_def['sub-options']
            else:
                logging.error(f'No parent-option for vendor {vendor}')
        else:
            logging.error(f'Vendor: {vendor} not defined')
        
        return sub_opt_defs
    
    
    def dump_vendor_def(self, vendor):
        '''
        Returns the vendor definition as a dict

        Parameters:
            vendor (str): Vendor Identifier
        
        Returns:
            dict containing vendor definition
        '''
        vendor_def = {}
        if self.included(vendor):
            vendor_def = self.config['vendors'][vendor]
        
        return vendor_def


# DHCP Decoding Utils

class dhcp_decode():
    '''
    Class to assist with Hex Encoding of 
    DHCP Options and sub_options
    '''
    def __init__(self) -> None:
        self.opt_types = [ 'string', 
                           'ip',
                           'array_of_ip',
                           'ipv4_address', 
                           'ipv6_address',
                           'boolean',
                           'int8',
                           'uint8',
                           'int16',
                           'uint16',
                           'int32',
                           'uint32',
                           'fqdn',
                           'binary',
                           'empty' ]

        self.fqdn_re, self.url_re = bloxone.utils.buildregex()
        return


    def hex_string_to_list(self, hex_string):
        '''
        Take a hex string and convert in to a list

        Parameters:
            hex_string (str): Hex represented as string

        Returns:
            list of hex bytes
        '''
        hex_list = []
        
        # Remove colons if present
        hex_string = hex_string.replace(':','')

        # Turn hex_string into a list
        for h in range(0, len(hex_string), 2):
            hex_list.append(hex_string[h:h+2])

        return hex_list

    def hex_to_suboptions(self, hex_string, encapsulated=False):
        '''
        Extract the sub-options from the hex data 
        '''
        hex_list = []
        index = 0
        subopt = {}
        suboptions = []
        opt_len = 0
        opt_data = ''
        opt_code = ''

        # Turn hex_string into a list
        hex_list = self.hex_string_to_list(hex_string)
            
        # If encapsulated assume first two bytes
        if encapsulated:
            index = 2

        while index <= (len(hex_list) - 2 ):
            opt_code = hex_list[index]
            opt_len = int(hex_list[index+1], 16)
            # Get option data
            for i in range(index + 2, (index + opt_len + 2)):
                if i >= len(hex_list):
                    logging.error(f'Data encoding error, non-standard format')
                    break    
                else:
                    opt_data += hex_list[i]
            
            # Build sub_opt and add to list of suboptions
            sub_opt = { 'code': self.hex_to_optcode(opt_code),
                        'data_length': opt_len,
                        'data': opt_data }
            
            suboptions.append(sub_opt)

            # Reset opt_data
            opt_data = ''

            # Move index 
            index = index + opt_len + 2
        
        return suboptions


    def validate_ip(self, ip):
        '''
        Validate input data is a valid IP address
        (Supports both IPv4 and IPv6)

        Parameters:
            ip (str): ip address as a string

        Returns:
            bool: Return True for valid and False otherwise

        '''
        try:
            ipaddress.ip_address(ip)
            result = True
        except ValueError:
            result = False
        return result


    # IP encondings
    def hex_to_ip(self, hex_string):
        '''
        Decode a 4 or 16 octect hex string to an IPv4 or IPv6 string 

        Parameters:
            hex_string (str): Hex representation of an IPv4 or IPv6 address 
        
        Returns:
            IP Address as a string
        '''
        ip = ''
        hex_string = hex_string.replace(':','')
        no_of_octects = self.hex_length(hex_string) 

        # Check number of octets
        if no_of_octects == '04' or no_of_octects == '10':
            # Assume IPv4 or IPv6
            int_ip = int(hex_string, 16)
            if self.validate_ip(int_ip):
                ip = ipaddress.ip_address(int_ip).exploded
            else:
                logging.error(f'{hex_string} not a valid IP Address')
                ip = ''
        else:
            ip = ''

        return ip

    def hex_to_array_of_ip(self, hex_string):
        '''
        Decode array of IPv4 or IPv6 addresses to CSV string 

        Parameters:
            hex_string (str): Hex representation of an array of IPv4 or IPv6
        
        Returns:
            IP Addresses in a CSV string
        
        '''
        array_of_ip = ''
        
        hex_length = int(self.hex_length(hex_string),16)

        if hex_length in [ 8, 12, 16, 20, 24 ]:
            # Assume IPv4
            for ip in [hex_string[n:n+8] for n in range(0, len(hex_string), 8)]:
                dip = self.hex_to_ip(ip)
                array_of_ip += dip + ','
            
        elif hex_length in [ 32, 48, 64 ]:
            # Assume IPv6
            for ip in [hex_string[n:n+32] for n in range(0, len(hex_string), 32)]:
                dip = self.hex_to_ip(ip)
                array_of_ip += dip + ','

        else:
            array_of_ip = 'array_of_ip_failed.'
            
        array_of_ip = array_of_ip[:-1]
            
        return array_of_ip


    # Methods for IPv4 and IPv6
    def hex_to_ipv4_address(self, hex_string):
        '''
        Decode a hex string to an IPv4 Address as a string

        Parameters:
            hex_string (str): Hex representation of an IPv4 address 
        
        Returns:
            IPv4 Address as a string
        '''
        hex_string = hex_string.replace(':','')
        return self.hex_to_ip(hex_string)
        
    
    def hex_to_ipv6_address(self, hex_string):
        '''
        Decode a hex string to an IPv6 address as a string 

        Parameters:
            hex_string (str): Hex representation of an IPv6 address 
        
        Returns:
            IPv6 Address as a string
        '''
        hex_string = hex_string.replace(':','')
        return self.hex_to_ip(hex_string)


    # String/text encoding
    def hex_to_string(self, hex_string):
        '''
        Decode a string of hex values to a text string

        Parameters:
            hex_string (str): Hex representation of a string
        
        Returns:
            text string (str)
        '''
        hex_string = hex_string.replace(':','')
        s = binascii.unhexlify(hex_string).decode()
        return s


    # Boolean encoding
    def hex_to_boolean(self, hex_string):
        ''' 
        Decode Hex value as a string to 'true' or 'false'

        Parameters:
            hex_string (str): Hex value as a str
        
        Returns:
           string representation of a boolean
        '''
        hex_string = hex_string.replace(':','')

        # Assume true if not zero i.e. check all bits for non-zero
        if int(hex_string, 16) == 0:
            text_bool = 'False'
        else:
            text_bool = 'True'
        
        return text_bool


   # integer encodings 
    def hex_to_int(self, hex_string, size=8):
        '''
        Decode hex to signed integer of defined size

        Parameters:
            hex_string (str): hex value as string
            size (int): size in bits [8, 16, 32]
        
        Returns:
            integer
        '''
        value = 0
        i_sizes = [ 8, 16, 32 ]
        hex_string = hex_string.replace(':','')

        i = int(hex_string, 16)
        if size in i_sizes:
            max_bits = size - 1
            if i < (2**size):
                if (i > (2**max_bits)):
                    value = -abs(i - (2**max_bits))
                else:
                    value = i
            else:
                raise ValueError(f'{i} is out of range for int{size} type')
        else:
            raise ValueError(f'Size must be 8, 16, or 32')

        return value 


    def hex_to_uint(self, hex_string, size=8):
        '''
        Encode integer of specified size as unsigned int in hex
        Uses 2's compliment if supplied with negative number


        Parameters:
            i (int): integer value to encode
            size (int): size in bits [8, 16, 32]
        
        Returns:
            hex encoding as string
        '''
        i_sizes = [ 8, 16, 32 ]
        hex_string = hex_string.replace(':','')
        i = int(hex_string, 16)

        if size in i_sizes:
            max_size = 2**size
            if i < max_size:
                value = i
            else:
                raise ValueError(f'{i} is out of range for uint{size} type')
        else:
            raise ValueError(f'Size must be 8, 16, or 32')

        return value 


    # Methods for intX and uintX
    def hex_to_int8(self, value):
        return self.hex_to_int(value, size=8)


    def hex_to_uint8(self, value):
        return self.hex_to_uint(value, size=8)


    def hex_to_int16(self, value):
        return self.hex_to_int(value, size=16)


    def hex_to_uint16(self, value):
        return self.hex_to_uint(value, size=16)


    def hex_to_int32(self, value):
        return self.hex_to_int(value, size=32)


    def hex_to_uint32(self, value):
        return self.hex_to_uint(value, size=32)


    # FDQN Encoding
    def hex_to_fqdn(self, hex_string):
        '''
        Decode RFC 1035 Section 3.1 formatted hexa to fqdn

        Parameters:
            hex_string (str): hex encoded fqdn

        Returns:
            fqdn as string
        '''
        hex_list = []
        index = 0
        fqdn = ''
        label_len = 0
        label = ''

        # Turn hex_string into a list
        hex_list = self.hex_string_to_list(hex_string)
            
        label_len = int(hex_list[index], 16)
        while label_len != 0:
            
            # Build label
            for i in range(index + 1, (index + label_len + 1)):
                label += hex_list[i]
            
            # Build fqdn and reset label
            fqdn += self.hex_to_string(label) + '.'
            label = ''

            # Reset index and check
            index = index + label_len + 1
            if index < len(hex_list):
                label_len = int(hex_list[index], 16)
            else:
                logging.warning('Reach end before null')
                label_len = 00

        return fqdn


    # Binary Encoding
    def hex_to_binary(self, data):
        '''
        Format hex string of binary/hex encoded data

        Parameters:
            data (str): data to format

        Returns:
            hex encoding as string
        '''
        hex_value = ''
        base = 16 

        # Check for binary
        if data[:2] == '0b':
            base = 2
        else:
            hex_string = data.replace(':','')

        # Force hex encoding without 0x using base
        hex_value = '{:02x}'.format(int(data, base))

        return hex_value

    
    # Empty Encoding
    def hex_to_empty(self, data):
        '''
        Return empyt hex string ''

        Parameters:
            data (str): Data not to encode (should be empty)
        
        Returns:
            Empty String ''
        '''
        if data:
            data = ''
        return data
        

    # Code and Length encoding
    def hex_to_optcode(self, hex_string):
        '''
        Encode Option Code in hex (1-octet)

        Parameters:
            optcode (str/int): Option Code 
        
        Returns:
            hex encoding as string
        '''
        opt_code = self.hex_to_int8(hex_string)
        return opt_code


    def hex_length(self, hex_string):
        '''
        Encode Option Length in hex (1-octet)

        Parameters:
            hex_string (str): Octet Encoded Hex String
        
        Returns:
            Number of Hex Octects as hex encoded string
        '''
        hex_string = hex_string.replace(':','')
        hex_len = '{:02x}'.format(int(len(hex_string) / 2))
        return hex_len


    def check_data_type(self, optcode, sub_defs=[]):
        '''
        Get data_type for optcode from sub optino definitions

        Parameters:
            optcode (int): Option code to check
            sub_defs (list of dict): sub option definitions to cross
                    reference
        
        Returns:
            data_type as str
        
        '''
        data_type = ''

        if sub_defs:
            for d in sub_defs:
                if int(optcode) == int(d['code']):
                    data_type = d['type']
                    # Check for array_of_ip 
                    if 'array' in d.keys():
                        if ('ip' in data_type) and d['array']:
                            data_type = 'array_of_ip'
                    break
        
        return data_type
 

    def get_name(self, optcode, sub_defs=[]):
        '''
        Get data_type for optcode from sub optino definitions

        Parameters:
            optcode (int): Option code to check
            sub_defs (list of dict): sub option definitions to cross
                    reference
        
        Returns:
            name as str
        
        '''
        name = ''

        if sub_defs:
            for d in sub_defs:
                if optcode == d['code']:
                    name = d['name']
                    break
        
        return name
                            

    def guess_data_type(self, subopt, padding=False):
        '''
        '''
        data_type = ''
        data_types = []
        dl = subopt['data_length']
        data = subopt['data'].replace(':','')

        # Check for 1 byte first
        if dl == 1:
            # int8 or bool (so treat as int8)
            # data_types.append('int8')
            data_type = 'int8'
        else:
            # We know it has more than one byte
                
            if data[-2:] == '00' and not padding: 
                # Possible FQDN
                logging.debug('Checking fqdn guess')
                fqdn = self.hex_to_fqdn(subopt['data'])
                # Validate FQDN
                if bloxone.utils.validate_fqdn(fqdn, self.fqdn_re):
                    logging.debug('FQDN verified')
                    data_types.append('fqdn')
                    data_type = 'fqdn'

            if dl in [4, 16]:
                logging.debug('CHecking for type ip')
                if self.hex_to_ip(data):
                    data_types.append('ip')
                    data_type = 'ip'

            if dl in [8, 32]:
                logging.debug('Checking for array of ip')
                if self.hex_to_ip(data[:dl]):
                    data_types.append('ip')
                    data_type = 'array_of_ip'
            
            if data_type == '':
                logging.debug('Default guess of string')
                data_type = 'string'            


        return data_type
        
    
    def decode_data(self, data, data_type='string', 
                    padding=False,
                    pad_bytes=1,
                    array=False):
        '''
        '''
        decoded = ''
        if data_type in self.opt_types:
            if 'ip' in data_type and array:
                data_type = 'array_of_ip'
            hex_to_type = eval('self.' + 'hex_to_' + data_type)
        else:
            logging.error(f'Unsupported Option Type {data_type}')
            logging.info('Unsupported option type, ' +
                         'attempting to process as string')
            hex_to_type = eval('self.hex_to_string')
        
        decoded = hex_to_type(data)

        return decoded
        

    def decode_dhcp_option(self, 
                           hex_string,
                           sub_opt_defs=[],
                           padding=False,
                           pad_bytes=1,
                           encapsulated=False,
                           id=None,
                           prefix=''):
        '''
        Attempt to decode DHCP options from hex representation

        Parameters:
            sub_opt_defs (list): List of Sub Option definition dictionaries
            padding (bool): Whether extra 'null' termination bytes are req.
            pad_bytes (int): Number of null bytes to append
            encapsulate (bool): Add id and total length as prefix
            id (int): option code to prepend
            prefix (str): String value to prepend to encoded options
        
        Returns:
            Encoded suboption as a hex string
        '''
        value = ''
        str_value = ''
        suboptions = []
        de_sub_opt = {}
        decoded_opts = []
        parent = {}
        guessed = False
        hex_string = hex_string.replace(':','')

        if (len(hex_string) % 2) == 0:
            if encapsulated:
                parent_opt = self.hex_to_opcode(hexstring[:2])
                total_len = self.hex_to_int8(hexstring[2:4])
                hex_string = hex_string[4:]
                parent = {'parent': parent_opt, 'total_len': total_len }
                decoded_opts.append(parent)

            # Break out sub-options
            suboptions = self.hex_to_suboptions(hex_string)

            # Attempt to decode sub_options
            for opt in suboptions:
                if sub_opt_defs:
                    data_type = self.check_data_type(opt['code'],
                                                    sub_defs=sub_opt_defs)
                    name = self.get_name(opt['code'], sub_defs=sub_opt_defs)

                else:
                    logging.debug(f'Attempting to guess option type for {opt}')
                    data_type = self.guess_data_type(opt)
                    guessed = True
                    name = 'Undefined'
                
                if data_type: 
                    value = self.decode_data(opt['data'], data_type=data_type)
                
                str_value = self.decode_data(opt['data'], data_type='string')

                de_sub_opt = { 'name': name,
                            'code': opt['code'],
                            'type': data_type,
                            'data_length': opt['data_length'],
                            'data': value,
                            'data_str': str_value,
                            'guess': guessed }

                decoded_opts.append(de_sub_opt)
        else:
            logging.error('Hex string contains incomplete octets')

        return decoded_opts


    def output_decoded_options(self, decoded_opts=[], output='pprint'):
        '''
        Simple output for decode_dhcp_options() data

        Parameters:
            decoded_opts (list): List of dict
            output (str): specify format [pprint, csv, yaml]
        
        '''
        formats = [ 'csv', 'pprint', 'yaml']
        header = ''
        head_printed = False
        line = ''
        
        if len(decoded_opts):
            if output in formats:
                # Output simply with pprint
                if output == 'pprint':
                    pprint(decoded_opts)
                # Output to CSV
                if output == 'csv':
                    for item in decoded_opts: 
                        if 'parent' in item.keys():
                            header = 'Parent, Total Length'
                            pprint(header)
                            pprint(f'{item["parent"]}, ' +
                                  f'{item["total_len"]}')
                        elif not head_printed:
                            header = ''
                            for key in item.keys():
                                header += key + ','
                            header = header[:-1]

                            pprint(header)
                            head_printed = True
                        else:
                            for key in item.keys():
                                line += repr(item[key]) + ','
                            line += line[:-1]
                            pprint(line)
                            line = ''
                # Output to normalised YAML
                if output == 'yaml':
                    try:
                        y = yaml.safe_dump(decoded_opts)
                        print(y)
                    except:
                        print('Could not convert to yaml')

            else:
                print(f'{output} not supported for output')
                print(f'Suported formats include: {formats}')
                print(decoded_opts)
        else:
            print('No option data')
        
        return
        
        
    def tests(self):
        '''
        Run through encoding methods and output example results
        '''
        encode = bloxone.dhcp_encode()
        test_data = [ { 'code': '1', 'type': 'string',
                        'data': 'AABBDDCCEEDD-aabbccddeeff' },
                      { 'code': '2', 'type': 'ipv4_address',
                        'data': '10.10.10.10' },
                      { 'code': '3', 'type': 'ipv4_address',
                        'data': '10.10.10.10,11.11.11.11', 'array': True },
                      { 'code': '4', 'type': 'boolean', 'data': True },
                      { 'code': '5', 'type': 'int8', 'data': '22' },
                      { 'code': '5', 'type': 'int8', 'data': '-22' },
                      { 'code': '6', 'type': 'uint8', 'data': '22' },
                      { 'code': '7', 'type': 'int16', 'data': '33'},
                      { 'code': '8', 'type': 'int16', 'data': '33'},
                      { 'code': '9', 'type': 'uint16', 'data': '33'}, 
                      { 'code': '10', 'type': 'int32', 'data': '44'}, 
                      { 'code': '11', 'type': 'uint32', 'data': '-44'}, 
                      { 'code': '12', 'type': 'uint32', 'data': '44'}, 
                      { 'code': '13', 'type': 'fqdn',
                        'data': 'www.infoblox.com' },
                      { 'code': '14', 'type': 'binary', 'data': 'DEADBEEF' },
                      { 'code': '15', 'type': 'empty', 'data': ''},
                      { 'code': '16', 'type': 'ipv6_address',
                        'data': '2001:DB8::1' },
                      { 'code': '17', 'type': 'ipv6_address',
                        'data': '2001:DB8::1,2001:DB8::2', 'array': True } ]

        print(f'Decoding types supported: {self.opt_types}')
        print()
        print('Non-array tests:')
        for data_test in test_data:
            enc_str = encode.encode_data(data_test)
            if 'array' in data_test.keys():
                array = data_test['array']
            else:
                array=False

            dec_str = self.decode_data(enc_str, 
                                       data_type=data_test['type'],
                                       array=array)
            print(f'Type: {data_test["type"]}, Hex: {enc_str}, ' +
                  f'Decoded: {dec_str}, Original: {data_test["data"]}')
        
        print()
        # Padding Test
        # test_data = { 'code': '99', 'type': 'string', 'data': 'AABBCCDD' }
        # result = encode.encode_data(test_data, padding=True)
       #  print(f'Padding test (1 byte), type string: {test_data["data"]}' +
        #       f' {result}')
        # Full encode test
        test_data = [ { 'code': '1', 'type': 'string',
                        'data': 'AABBDDCCEEDD-aabbccddeeff' },
                      { 'code': '2', 'type': 'ipv4_address',
                        'data': '10.10.10.10' },
                      { 'code': '3', 'type': 'ipv4_address',
                        'data': '10.10.10.10,11.11.11.11', 'array': True },
                      { 'code': '4', 'type': 'boolean', 'data': True },
                      { 'code': '5', 'type': 'int8', 'data': '22' } ]
        result = encode.encode_dhcp_option(test_data)
        print(f'Full encoding of sample Hex: {result}')
        decode = self.decode_dhcp_option(result, sub_opt_defs=test_data)
        print(f'Decoding result:')
        self.output_decoded_options(decode)

        return


#!/usr/bin/env python3
#vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''

 Description:

    BloxOne b1user class providing an simple data interface for user info 

 Requirements:
   Python3 with re, ipaddress, requests and sqlite3 modules

 Author: Chris Marrison

 Date Last Updated: 20230127

 Todo:

 Copyright (c) 2023 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

'''
__version__ = '0.1.0'
__author__ = 'Chris Marrison'
__author_email__ = 'chris@infoblox.com'

import bloxone
import logging


class b1user(bloxone.b1platform):
    '''
    '''
    def __init__(self,
                 username: str = '', 
                 email_domain: str = 'infoblox.com',
                 b1org: str = '', 
                 delete: bool = False, 
                 cfg_file='config.ini'):
        '''
        Initialise class

        Parameters:
            username (str): Username or email address
            email_domain (str): email domain of user
            b1org (str): Name of org 
            delete (bool): Delete data and user or report only
            cfgfile (str): inifile to user for bloxone module
        '''
        super().__init__(cfg_file)

        self.username = username
        self.email_domain = email_domain
        self.b1org = b1org
        self.delete = delete
        
        user_dict = self.convert_username()
        self.owner = user_dict.get('owner')
        self.email = user_dict.get('email')

        self.current_org = self.get_current_tenant()

        return


    def convert_username(self) -> dict:
        '''
        '''
        user_dict = {}
        u = self.username.casefold()
        if '@' in self.username:
            user = u.split('@')[0]
            email = self.username
        else:
            user = u
            email = f'{u}@{self.email_domain}'
        
        user_dict = { 'owner': user, 'email': email }

        return user_dict


    def check_tenant(self) -> bool:
        '''
        Check Org
        '''
        return self.current_org == self.b1org


    def user_exists(self) -> bool:
        '''
        '''
        status = False
        user_id = self.get_user_id(self.email)
        if user_id:
            status = True
        else:
            status = False

        return status


    def is_current_user(self) -> bool:
        '''
        '''
        status = False
        response = self.get_current_user()
        if response.status_code in self.return_codes_ok:
            current_user = response.json().get('result').get('email')
            if self.email == current_user:
                status = True
            else:
                status = False
        else:
            logging.error('Could not determine current user.')
            logging.error(f'HTTP code: {response.status_code} ')
            logging.error(f'Response: {response.text}')
            raise

        return status

        
    def find_user_objects(self) -> dict:
        '''
        '''
        return self.simple_global_search(self.owner)


    def find_join_tokens(self) -> dict:
        '''
        '''
        filter = f'name~"{self.username}*."'
        response = self.get_join_token(_filter=filter)
        if response.status_code in self.return_codes_ok:
            result = response.json().get('results')
        else:
            logging.error(f'HTTP code: {response.status_code} ')
            logging.error(f'Response: {response.text}')
            result = {}

        return result


    def find_ophs(self) -> dict:
        '''
        '''
        filter = f'display_name~"{self.owner}*."'
        response = self.on_prem_hosts(_filter=filter)
        if response.status_code in self.return_codes_ok:
            result = response.json().get('result')
        else:
            logging.error(f'HTTP code: {response.status_code} ')
            logging.error(f'Response: {response.text}')
            result = {}

        return result


    def data_report(self) -> bool:
        '''
        '''
        status = False
        lines = []
        search = self.find_user_objects()
        jts = self.find_join_tokens()
        ophs = self.find_ophs()

        if search or jts or ophs:
            # We found something
            status = True
            if search:
                print('Data objects:\n')
                for item in search:
                    lines.append( f'Name: {item.get("name")}, ' +
                                  f'Resource: {item.get("metadata").get("resource")}, ' +
                                  f'Tags: {item.get("tags")}' )
                for l in lines:
                    print(l)
            else:
                print('No data objects found.')

            if jts:
                lines = []
                for item in jts:
                    lines.append( f'Name: {item.get("name")}, ' +
                                  f'Status: {item.get("status")}, ' +
                                  f'Use-counter: {item.get("use_counter")}' +
                                  f'Tags: {item.get("tags")}' )
                # Print results
                print('\nJoin Tokens:\n')
                for l in lines:
                    print(l)
            else:
                print('No join tokens found.')
            
            if ophs:
                lines = []
                for item in ophs:
                    lines.append( f'Name: {item.get("display_name")}, ' +
                                  f'Last seen: {item.get("last_seen")}, ' +
                                  f'Tags: {item.get("tags")}' )
                # Print results
                print('\nOn Prem Hosts:\n')
                for l in lines:
                    print(l)
            else:
                print('No OPHs found.')

        else:
            # Nothing found
            status = False

        return status

"""
            # Check current user
            if delete:
                # Check current user
                response = b1p.get_current_user()
                if response.status_code in b1p.return_codes_ok:
                    current_user = response.json().get('email')
                    if current_user == user['email']:
                        logging.error('Attempting to delete current user!')
                        exitcode = 1
                        raise exit(exitcode)
                else:
                    logging.error('Could not determine current user.')
                    logging.error(f'Response code: {response.status_code}')
                    logging.error(f'Message: {response.text}')
                    exitcode = 2
                    raise exit(exitcode)
            
            # Collect

                
            '''
            Input: Username and Org
            - Check for correct Org
            - Determine username
                - Does user exist
            - Find objects
            - Process object list
                - Determine Objecname and Owner tags
                - Check against username
                - Report
            - Delete process
                - order of deletion
                - delete objects

            '''

        
        else:
            logging.error(f'Running against wrong BloxOne Org.')
            logging.info(f'Current Org: {org}')
            exitcode = 1

        return status
"""

### Main ###
if __name__ == '__main__':
    exit(1)
## End Main ###


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 Module to provide class hierachy to simplify access to the BloxOne APIs

 Date Last Updated: 20210713

 Todo:

 Copyright (c) 2021 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
from bloxone.b1oph import b1oph
import bloxone
import logging
import requests
import json

# ** Global Vars **
__version__ = '0.1.0'
__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__doc__ = 'https://python-bloxone.readthedocs.io/en/latest/'
__license__ = 'BSD'


class APIError_Unable_To_Retrieve_Commands(Exception):
    '''
    Exception for API call in __init__
    '''
    pass


class Command_Not_Supported(Exception):
    '''
    Exception for API call in __init__
    '''
    pass


class Unknown_Argument(Exception):
    '''
    Exception for API call in __init__
    '''
    pass


class b1diagnostics(bloxone.b1):
    '''
    Class to simplify access to the BloxOne Platform APIs
    '''

    def __init__(self, cfg_file='config.ini'):
        '''
        Call base __init__ and extend
        '''
        super().__init__(cfg_file)

        # Instantiate b1oph class as need access to b1oph.get_ophid()
        self.b1_oph = b1oph(cfg_file)

        # Automatically get list of remote_commands
        try:
            response = self.get_remote_commands()
            self.commands = response.json()['results']
        except:
            logging.error(f'Response code: {response.status_code}')
            logging.error(f'Response body: {response.text}')
            raise APIError_Unable_To_Retrieve_Commands()
        
        return


    def get(self, objpath, id="", action="", **params):
        '''
        Generic get object wrapper for platform calls

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"

        Returns:
            response object: Requests response object
        '''

        # Build url
        url = self.diagnostics_url + objpath
        url = self._use_obj_id(url,id=id)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response

        
    def post(self, objpath, body=""):
        '''
        Generic create object wrapper for platform objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.diagnostics_url + objpath
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, objpath, id=""):
        '''
        Generic delete object wrapper for platform objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to delete

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.diagnostics_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apidelete(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.diagnostics_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response


    # *** Helper Methods ***

    def get_id(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id = ""
        filter = key+'=="'+value+'"'
        fields = key + ',id'

        # Make API Call
        response = self.get(objpath, _filter=filter, _fields=fields)

        # Process response
        if response.status_code in self.return_codes_ok:
            obj = response.json()
            # Look for results
            if "results" in obj.keys():
                obj = obj['results']
                if obj:
                    id = obj[0]['id']
                    if not include_path:
                        id = id.rsplit('/',1)[1]
                else:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id

    # Helper Methods

    '''
    def get_ophid(self):
        return ophid
    '''

    def get_remote_commands(self):
        '''
        Get set of possible remote commands and parameters

        Returns:
            response object: Requests response object
        '''
        return self.get('/remotecommands')


    def is_command(self, command):
        '''
        Check whether command is valid

        Parameters:
            command(str): command to check
        
        Returns:
            boolean
        '''
        cmds = []
        for cmd in self.commands:
            cmds.append(cmd['name'])
        if command in cmds:
            status = True
        else:
            status = False
        return status

    
    def get_args(self, command):
        '''
        Check the args for a command

        Parameters:
            command(str): Command to retrieve argyments for

        Returns:
            Disctionary of arguments or empty dictionary if none.
        
        Raises:
            Command_Not_Supported Exception if command is not available
        '''
        args = {}
        if self.is_command(command):
            for cmd in self.commands:
                if cmd['name'] == command:
                    if 'args' in cmd.keys():
                        args = cmd['args']
                    else:
                        args = {}
                    break
        else:
            raise Command_Not_Supported(f'Command: {command} not supported.')
        
        return args



    def execute_task(self, command, args={}, ophname=None, 
                     ophid=None, id_only=True, priv=False):
        '''
        Execute remote command on an OPH
    
        Parameters:
            cmd(str): Command to execute
            args(dict): Command arguments
            ophname(str): Name of OPH to execute command on (or supply ophid)
            ophid(str): (Optional) ophid of OPH for cmd execution
            id_only(bool): default of True
            priv(bool): Run privileged task, default of False

        Returns:
            id string of task if id_only=True (defult)
            response object: Requests response object if id_only=False

        Raises:
            TypeError Exception if required options not supplied 
            KeyErro Exception if ophname is not found (and ophid not supplied)
            Command_Not_Supported Exception if command is not valid
            Unknown_Argument Exception if arguments do not match required
        
        Todo:
            [ ] Enhance logic to run /priviledgetask or /task
            Awaiting API enhancement to determine priv versus non-priv
            [ ] Enhance args check for required arguments
            Awaiting API enhancement for arg to determine required versus 
            optional arguments
        '''
        # Check command is valid
        if self.is_command(command):
            # If ophid supplied then get this
            if ophid or ophname: 
                if not ophid:
                    logging.debug(f'Getting ophid for OPH: {ophname}')
                    ophid = self.b1_oph.get_ophid(name=ophname)
                    if not ophid:
                        logging.error(f'OPH not found.')
                        raise KeyError(f'OPH {ophname} not found')
                logging.debug(f'OPHID: {ophid}')

                # Check args
                arglist = self.get_args(command)
                for arg in args.keys():
                    if arg not in arglist.keys():
                        raise Unknown_Argument(f'Argument: {arg} not recognised')
                    s_type = type(args[arg]) 
                    ex_type = type(arglist[arg])
                    if s_type != ex_type:
                        raise TypeError(f'Supplied argument type {s_type}' +
                            f' does not match expected type {ex_type}')

                # Create command body
                body = '{ "cmd": '
                if len(args) > 0:
                    body = body + '{ "args": ' + json.dumps(args) + ', '
                else:
                    body = body + '{ '
                body = body + '"name": "' + command + '" }, '
                body = body + '"ophid": "' + ophid + '" }'
                logging.debug(f'Task body: {body}')

                # Execute
                if priv:
                    response = self.post('/privilegedtask', body=body)
                else:
                    response = self.post('/task', body=body)
                if id_only:
                    if response.status_code in self.return_codes_ok:
                        id = response.json()['result']['id']
                    else:
                        error_msg = ( 'Failed to create task, ' +
                                    f'HTTP Code: {response.status_code}')
                        logging.error(error_msg)
                        id = None
                    result = id
                else:
                    result = response

            else:
                logging.error('No ophname or ophid supplied.')
                raise TypeError('Requires either ophname or ophid to be defined')
        else:
            logging.error(f'Command: {command} is not a supported command')
            raise Command_Not_Supported(f'Command: {command} not supported.')
        
        return result


    def get_task_result(self, taskid):
        '''
        Get the results for specidied task

        Parameters:
            taskid(str): id of executed task
        
        Returns:
            response object: Requests response object if id_only=False
        '''
        id = '/task/' + taskid
        return self.get(id)
        

    def download_task_results(self, taskid):
        '''
        Get the results for specidied task

        Parameters:
            taskid(str): id of executed task
        
        Returns:
            response object: Requests response object if id_only=False

        Note:

        '''
        id = '/task/' + taskid + '/download'
        return self.get(id)
        


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 BloxOne Threat Defense Enpoint API Wrapper Class and Helpers

 Author: Chris Marrison

 Date Last Updated: 20240201

 Todo:

 Copyright (c) 2020 - 2024 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.3.0'
__author__ = 'Chris Marrison/Krishna Vasudeva'
__author_email__ = 'chris@infoblox.com'

import bloxone
import logging
import csv
import sys

# ** Global Vars **

class b1tdlad(bloxone.b1td):

    # Generic Methods
    def get(self, objpath, **params):
        '''
        Generic get object wrapper for Threat Defense objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.tdlad_url + objpath
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper 

        Parameters:
            objpath (str):  Swagger object path
            body (str): JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url if needed
        url = self.tdlad_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response
    

    def get_lookalikes(self, target_domain='', **params):
        '''
        '''
        url = f'{self.tdlad_url}/lookalikes'

        if target_domain:
            filter = f'target_domain=="{target_domain}"'
            url = self._add_params(url, _filter=filter)
            url = self._add_params(url, first_param=False, **params)
        else:
            url = self._add_params(url, **params)
        
        response = self._apiget(url)
        
        return response
    

    def lookalike_to_csv(self, records, file=sys.stdout, enriched=False):
        '''
        Output lookalike data to CSV

        Parameters:
            records (list): Lookalike API result data
            file (filehandler): File opened for write with newline=''
            enriched (bool): Enhance data using dossier
        '''
        lookalikes = records.get('results')
        if not lookalikes:
            lookalikes = records
        
        if lookalikes:
            # Check whether this is enriched output
            if enriched:
                headers = ['target_domain', 'lookalike_domain', 
                           'registration_date', 'phishing', 'malware', 
                           'suspicious', 'categories', 'reason', 'registrar',
                           'last_updated', 'country', 'region',
                           'cert_issued_by', 'cert_expiry' ]
                w = csv.DictWriter(file, fieldnames=headers, extrasaction='ignore')
                w.writeheader()
                for l in lookalikes:
                    # Enrich lookalike 
                    r = self.enrich_lookalike(l)
                    w.writerow(l)

            else:
                # Simple Output
                headers = ['target_domain', 'lookalike_domain', 'registration_date', 
                        'phishing', 'malware', 'suspicious', 'categories', 'reason']

                try:
                    w = csv.DictWriter(file, fieldnames=headers, extrasaction='ignore')
                    w.writeheader()
                    w.writerows(lookalikes)
                except Exception as err:
                    logging.error(f'Error occured writing CSV: {err}')
                    raise
        else:
            logging.error('No data found')

        return
        

    def enrich_lookalike(self, record):
        '''
        '''
        sources = [ "geo", "whois", "ssl_cert" ]
        host = record.get('lookalike_domain')

        if host:
            response = self.dossierquery(host, type='host', sources=sources )
            if response.status_code in self.return_codes_ok:
                results = response.json().get('results')
                if results:
                    for r in results:
                        src = r.get('params').get('source')
                        data = r.get('data')
                        # Check data source
                        if src == 'whois':
                            if data.get('response'):
                                pw = data.get('response').get('parsed_whois')
                                if pw:
                                    registrar = pw.get('registrar').get('name')
                                    udate = pw.get('updated_date')
                                else:
                                    registrar = None
                                    udate = None
                            else:
                                registrar = None
                                udate = None
                        elif src == 'geo':
                            country = data.get('country_name')
                            region = data.get('region')
                        elif src == "ssl_cert":
                            ci = data.get('issuer')
                            ssl_chain = data.get('ssl_cert_chain')
                            if ssl_chain:
                                cert_exp = ssl_chain[0].get('expires')
                            else:
                                cert_exp = None

                    # Update record
                    enrichment = { 'registrar': registrar,
                                   'last_updated': udate,
                                   'country': country,
                                   'region': region,
                                   'cert_issued_by': ci,
                                   'cert_expiry': cert_exp }
                    
                    record.update(enrichment)
            else:
                logging.error(f'Could not enrich, API error: {response.status_code}')
        else:
            logging.error('No lookalike found in record')

        return record

"""
    def obfiscate_domains(self, data: list, fields: list):
        '''
        '''

        return
"""


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 Module to provide class hierachy to simplify access to the BloxOne APIs

 Date Last Updated: 20230608

 Todo:

    Complete the service control methods

 Copyright (c) 2021-2023 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
import bloxone
import logging
import json

# ** Global Vars **
__version__ = '0.0.6'
__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__doc__ = 'https://python-bloxone.readthedocs.io/en/latest/'
__license__ = 'BSD'


class b1infra(bloxone.b1):
    '''
    Class to simplify access to the BloxOne Platform APIs
    This class replaces the deprecated b1oph class
    '''

    def __init__(self, cfg_file='config.ini'):
        '''
        Call base __init__ and extend
        '''
        super().__init__(cfg_file)

        self.b1host_COMPOSITE_STATUS = { '0': 'Review Details', 
                                      '1': 'Online',
                                      '2': 'Unknown',
                                      '3': 'Pending',
                                      '4': 'Awaiting serviceroval' }
        
        self.b1host_PLATFORM_STATES = { '0': 'Offline',
                                     '1': 'Online',
                                     '2': 'Error',
                                     '3': 'Waiting (pending)',
                                     '4': 'Unknown' }
        
        self.b1host_HOST_TYPES = { '0': 'Not Available',
                                '1': 'Unknown',
                                '2': 'Unknown',
                                '3': 'BloxOne VM',
                                '4': 'BloxOne serviceliance - B105',
                                '5': 'BloxOne Container',
                                '6': 'CNIOS' }
        
        self.b1host_service_MGT = { '0': 'Inactive',
                             '1': 'Active',
                             '2': 'Error',
                             '3': '' }

        self.b1host_services = { '1': { 'serviceName': 'DFP', 'StatusSpace': '9' },
                          '2': { 'serviceName': 'DNS', 'StatusSpace': '12' },
                          '3': { 'serviceName': 'DHCP', 'StatusSpace': '15' },
                          '7': { 'serviceName': 'CDC', 'StatusSpace': '24' },
                          '9': { 'serviceName': 'Anycast', 'StatusSpace': '30' },
                          '10': { 'serviceName': 'NGC', 'StatusSpace': '34' },
                          '12': { 'serviceName': 'MS AD Collector', 'StatusSpace': '40' },
                          '13': { 'serviceName': 'Access Authentication', 'StatusSpace': '43' },
                          '14': { 'serviceName': 'Edge_Services_FW', 'StatusSpace': '46' },
                          '15': { 'serviceName': 'Edge_Services_Router', 'StatusSpace': '49' },
                          '16': { 'serviceName': 'Site-to-Site_VPN', 'StatusSpace': '52' },
                          '18': { 'serviceName': 'DNS_Assured_Forwarding', 'StatusSpace': '58' },
                          '20': { 'serviceName': 'NTP', 'StatusSpace': '64' }
        }

        self.b1host_service_NAMES = { 'DFP': '1',
                               'DNS': '2',
                               'DHCP': '3',
                               'CDC': '7',
                               'Anycast': '9',
                               'NGC': '10' }

        self.service_STATUS = { '0': 'inactive', '1': 'active', '2': 'stopped' }

        return


    def get(self, objpath, id="", action="", **params):
        '''
        Generic get object wrserviceer for platform calls

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"

        Returns:
            response object: Requests response object
        '''

        # Build url
        url = self.infra_url + objpath
        url = self._use_obj_id(url,id=id)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response

        
    def create(self, objpath, body=""):
        '''
        Generic create object wrserviceer for platform objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.infra_url + objpath
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, objpath, id=""):
        '''
        Generic delete object wrserviceer for platform objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to delete

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.infra_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apidelete(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrserviceer for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.infra_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response


    def patch(self, objpath, id="", body=""):
        '''
        Generic create object wrserviceer for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.infra_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipatch(url, body)

        return response


    def get_tags(self, objpath, id=""):
        '''
        Get tags for an object id

        Parameters:
            objpath (str):  Swagger object path

            id (str): id of object

        Returns:
            tags (dict): Dictionary of current tags
                         or empty dict if none
        
        .. todo::
            * make generic, however, this requires the below
            * Lookup dictionary of 'required fields' per object type
        '''
        tags = {}
        response = self.get(objpath, id=id, _fields="tags")
        if response.status_code in self.return_codes_ok:
            tags = json.loads(response.text)
            tags = tags['result']
        else:
            logging.debug(f'HTTP Error occured. {response.status_code}')
            logging.debug(f'Body: {response.content}')
            tags = {}
        
        return tags


    # *** B1 Host/Service API Requests *** 

    def b1_hosts(self, **params):
        '''
        Method to retrieve BloxOne Hosts

        Parameters:
            **params (dict): Generic API parameters

        Returns:
            response object: Requests response object
        '''

        # Call BloxOne API
        response = self.get('/hosts', **params)

        # Return response object
        return response 


    def b1_detail_hosts(self, **params):
        '''
        Method to retrieve BloxOne Hosts
        Note: Undocumented and subject to change

        Parameters:
            **params (dict): Generic API parameters

        Returns:
            response object: Requests response object
        '''

        # Call BloxOne API
        response = self.get('/detail_hosts', **params)

        # Return response object
        return response 


    def get_id(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id = ""
        filter = key+'=="'+value+'"'
        fields = key + ',id'

        # Make API Call
        response = self.get(objpath, _filter=filter, _fields=fields)

        # Process response
        if response.status_code in self.return_codes_ok:
            obj = response.json()
            # Look for results or result
            if "results" in obj.keys():
                result_text = "results"
            elif "result" in obj.keys():
                result_text = "result"
            else:
                result_text = ""
            
            # Get result if available
            if result_text:
                obj = obj[result_text]
                if obj:
                    id = obj[0]['id']
                    if not include_path and "/" in str(id):
                        id = id.rsplit('/',1)[1]
                else:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


    def get_ophid(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):       ophid or ""
        '''

        # Local Variables
        id = ""
        filter = key+'=="'+value+'"'
        fields = key + ',ophid'

        # Make API Call
        response = self.get(objpath, _filter=filter, _fields=fields)

        # Process response
        if response.status_code in self.return_codes_ok:
            obj = response.json()
            # Look for results or result
            if "results" in obj.keys():
                result_text = "results"
            elif "result" in obj.keys():
                result_text = "result"
            else:
                result_text = ""
            
            # Get result if available
            if result_text:
                obj = obj[result_text]
                if obj:
                    id = obj[0]['ophid']
                    if not include_path and "/" in str(id):
                        id = id.rsplit('/',1)[1]
                else:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


    def get_b1host_id(self, name="", ophid=False):
        '''
        Return the id of named BloxOne Host

        Parameters:
            name (str): display name of b1host
        
        Returns:
            b1hostid(str): b1hostid of the specified b1host
        '''
        if ophid:
            id = self.get_ophid('/hosts', key="display_name", value=name, include_path=False)
        else:
            id = self.get_id('/hosts', key="display_name", value=name, include_path=False)
        
        return id


    def b1host_add_tag(self, id="", tagname="", tagvalue=""):
        '''
        Method to add a tag to an existing B1 Host

        Parameters:
            objpath (str):  Swagger object path
            tagname (str): Name of tag to add
            tagvalue (str): Value to associate

        Returns:
            response object: Requests response object
        '''
        # tags = self.get_tags('/hosts', id=id)
        response = self.get('/hosts', id=id, _fields="display_name,tags")
        if response.status_code in self.return_codes_ok:
            data = response.json()['result']
        else:
            data = {}
        logging.debug("Existing tags: {}".format(data))
        # Add new tag to data
        if tagname:
            data['tags'].update({tagname: tagvalue})
            logging.debug("New tags: {}".format(data))
        # Update object
        response = self.update('/hosts', id=id, body=json.dumps(data))

        return response


    def b1host_delete_tag(self, id="", tagname=""):
        '''
        Method to delete a tag from an existing B1 Host

        Parameters:
            objpath (str):  Swagger object path
            tagname (str): Name of tag to add

        Returns:
            response object: Requests response object
        '''
        # tags = self.get_tags('/hosts', id=id)
        response = self.get('/hosts', id=id, _fields="display_name,tags")
        if response.status_code in self.return_codes_ok:
            data = response.json()['result']
            logging.debug("Existing tags: {}".format(data))
            # Delete tag from data
            if tagname in data['tags'].keys():
                data['tags'].pop(tagname, True)
                print(json.dumps(data))
                logging.debug("New tags: {}".format(data))
                # Update object
                response = self.update('/hosts', id=id, body=json.dumps(data))
        else:
            logging.debug(f'HTTP Error occured. {response.status_code}')
            logging.debug(f'Body: {response.content}')

        return response


    def b1host_status_summary(self, name="", id="", tags=False):
        '''
        Get B1 Hosts status information for one or more Hosts

        Parameters:
            name (ste): Display name of Host
            id (str): id of a specific Host
            tags (bool): include tags in report
        
        Returns:
            rpt: Dict of translated status elements
        '''
        rpt = None

        # Get b1host Data
        if name:
            name_filter = 'display_name=="' + name + '"'
            response = self.get('/detail_hosts', _filter=name_filter)
        elif id:
            response = self.get('/detail_hosts', id=id)
        else:
            response = self.get('/detail_hosts')
        
        if response.status_code in self.return_codes_ok:
            rpt = self.b1host_status_rpt(response.json(), tags=tags)
        else:
            logging.error(f'Response code: {response.status_code}, Data: {response.text}')
            rpt = None
    
        return rpt


    def b1host_status(self, b1host_data):
        '''
        Translate status info in JSON data for an b1host

        Parameters:
            b1host_data (dict): Data for individual b1host
        
        Returns:
            b1host_status: Dict of translated status elements
        '''
        b1host_status = {}
        b1host_comp_state = ''
        b1host_configs = []
        
        b1host_comp_state = b1host_data.get('composite_status')
        if not b1host_comp_state:
            b1host_comp_state = 'No state information'
                        
        if 'configs' in b1host_data.keys():
            b1host_configs = b1host_data.get('configs')

        b1host_status = { 'b1host state': b1host_comp_state,
                          'configs': b1host_configs } 
        
        return b1host_status
    

    def b1host_service_status(self, b1host_data):
        '''
        Translate service status info in JSON data for an b1host

        Parameters:
            b1host_data (dict): Data for individual b1host
        
        Returns:
            b1host_services: Dict of translated status elements
        '''
        b1host_services = {}
        service_name = ''
        service_type = ''
        status = ''
        message = ''
        service_version = None

        # Check for servicelication status data
        if 'services' in b1host_data.keys():
            for service in b1host_data['services']:
                service_name = service.get('service_name')
                service_type = service.get('service_type')
                service_version = service.get('current_version')

                # Check whether service is enabled
                if 'status' in service.keys():
                    status = service['status'].get('status')
                    message = service['status'].get('message')


                # Add version info if available
                if message:
                    b1host_services.update( { service_name: { 
                                                'service_type': service_type,
                                                'status': status,
                                                'message': message,
                                                'version': service_version } 
                                            } )
                else:
                    b1host_services.update( { service_name: { 
                                                'server_type': service_type,
                                                'status': status,
                                                'version': service_version }
                                            } )

        else:
            b1host_services = {}
        
        return b1host_services
   

    def b1host_status_rpt(self, json_data, tags=False):
        '''
        Build  status report from GET /detail_hosts data
        Note: Undocumented API call use with caution

        Parameters:
            json_data (json): API JSON data for B1 Hosts call
            tags (bool): Include tags in response, default False
        
        Returns:
            rpt: Dict of status elements
        '''
        rpt = {}
        b1host_name = ""
        b1host_status = {}
        b1host_services = {}

        if json_data:
            results = json_data.get('results')
            if not isinstance(results, list):
                results = [ results ]
            # Build report 
            for b1host in results:
                b1host_services = {}
                b1host_name = b1host.get('display_name')
                if b1host_name:
                    b1host_status = self.b1host_status(b1host)
                    b1host_services = self.b1host_service_status(b1host)

                    rpt.update( { b1host_name: { 'status': b1host_status,
                                          'id': b1host.get('id'),
                                          'host_type': b1host.get('host_type'),
                                          'ip_address': b1host.get('ip_address'), 
                                          'pool': b1host.get('pool'),
                                          'nat_ip': b1host.get('nat_ip'),
                                          'version': b1host.get('current_version'),
                                          'last_seen': b1host.get('last_seen'),
                                          'services': b1host_services }
                                } )
                if tags:
                    rpt[b1host_name].update({'tags': b1host.get('tags')})
        else:
            rpt = None

        return rpt

    
    def b1host_uptime(self, name="", ophid=True):
        '''
        Get the uptime for a b1host

        Note: this currently uses the ophid, if set to False this may
              generate an error

        Parameters:
            name (str): Display Name of b1host
            ophid (bool): Defaults to True to use the ophid rather than id
        
        Returns:
            Number of seconds as a string
        '''
        uptime = None
        b1hostid = self.get_b1host_id(name=name, ophid=ophid)
        url = 'https://csp.infoblox.com/atlas-status-service/v1/getsummary'
        if ophid:
            body = '{"objectID":["' + b1hostid +'"],"objectType":["Onprem Host ID"],"event_key":["health-collector/heartbeat/"]}'
        else:
            body = '{"objectID":["' + b1hostid +'"],"objectType":["Host ID"],"event_key":["health-collector/heartbeat/"]}'
        response = self._apipost(url, body=body)
        if response.status_code in self.return_codes_ok:
            uptime = response.json()['results'][0]['metadata']['hostUptime']
        
        return uptime


    def get_service_state(self, service_name=''):
        '''
        Get status of a service

        Parameters:
            service (str): Service Name
        
        Returns:
            requests response object
        '''
        service_status = ''

		# Filter on service_name if provided
        if service_name:
            filter = f'name=="{service_name}"'
            logging.debug(f'Applying filter: {filter}')
            response = self.get('/services', _filter=filter)
        else:
            response = self.get('/services')

        # Check repsone
        if response.status_code in self.return_codes_ok:
            service_status = response.json()
        else:
            logging.debug(f'HTTP Error occured. {response.status_code}')
            logging.debug(f'Body: {response.content}')
        
        return service_status


    def manage_service(self, service_name="", action="status"):
        '''
        Perform action on named b1host for specified service

        Parameters:
            service_name (str): service Name
            action (str): action to perform for service
        
        Returns:
            bool
        '''
        result = False
        actions = [ 'status' ]
        # actions = [ 'status', 'disable', 'enable', 'stop', 'start' ]

        if action in actions:
           if action == "status":
                response = self.get_service_state(service_name=service_name)
                if response:
                    result = response
                else:
                    logging.error(f'Service: {service_name} not found')
                    result = False
           elif action == "disable":
               result = self.disable_service(service_name=service_name)
           elif action == "enable":
               result = self.enable_service(service_name=service_name)
           elif action == "start":
               result = self.service_process_control(service_name=service_name,
                                                     action="start") 
           elif action == "stop":
               result = self.service_process_control(service_name=service_name, 
                                                     action="stop") 
           else:
               logging.error(f'Action: {action} not implemented')
               result = False
        else:
            logging.error(f'Action: {action} not supported')
            logging.info(f'Supported actions: {actions}')
            result = False

        return result

    """

    def disable_service(self, service_name=''):
        '''
        Disable specified service on named b1host

        Parameters:
            name (str): display_name of b1host
            service (str): service Name, e.g. DNS
        
        Returns:
            bool
        '''
        status = False
        service_type = ''

        # Check service supported and get Get servicelication_type
        if service_name in self.b1host_service_NAMES.keys():
            service_type = self.b1host_service_NAMES[service]
        elif service_name in self.b1host_services.keys():
            service_type = service

        if service_type:
            # Get id of b1host
            filter = f'display_name=="{name}"'
            response = self.get('/hosts', 
                                _filter=filter)
            if response.status_code in self.return_codes_ok:
                b1host_data = response.json()['result']
                id = b1host_data[0]['id']
                logging.debug(f'B1 Host id = {id}')

                # Build body
                body = { 'display_name': name, 
                         'servicelications': [ { 'servicelication_type': service_type,
                                             'disabled': '1', 
                                             'state': { 'desired_state': '0' } 
                                           } ] } 
                
                # Update desired b1host
                response = self.update('/hosts',
                                       id=id,
                                       body=json.dumps(body))
                if response.status_code in self.return_codes_ok:
                    logging.debug(f'b1host: {name}, service: {service}, service_type: {service_type}')
                    status = True
                else:
                    logging.error(f'{response.status_code}: {response.text}')

            else:
                logging.error(f'b1host {name} not found.')
                status = False
        else:
            logging.error(f'service {service} not supported.')
            status = False
        
        return status


    def enable_service(self, name="", service=""):
        '''
        Enable specified service on named b1host

        Parameters:
            name (str): display_name of b1host
            service (str): service Name, e.g. DNS
        
        Returns:
            bool
        '''
        status = False
        service_type = ''

        # Check service supported and get Get servicelication_type
        if service in self.b1host_service_NAMES.keys():
            service_type = self.b1host_service_NAMES[service]
        elif service in self.b1host_services.keys():
            service_type = service

        if service_type:
            # Get id of b1host
            filter = f'display_name=="{name}"'
            response = self.get('/hosts', 
                                _filter=filter)
            if response.status_code in self.return_codes_ok:
                b1host_data = response.json()['result']
                id = b1host_data[0]['id']
                logging.debug(f'B1 Host id = {id}')

                # Build body
                body = { 'display_name': name, 
                         'servicelications': [ { 'servicelication_type': service_type,
                                             'disabled': '0', 
                                             'state': { 'desired_state': '1' } 
                                           } ] } 
                
                # Update desired b1host
                response = self.update('/hosts',
                                       id=id,
                                       body=json.dumps(body))
                if response.status_code in self.return_codes_ok:
                    logging.debug(f'b1host: {name}, service: {service}, service_type: {service_type}')
                    status = True
                else:
                    logging.error(f'{response.status_code}: {response.text}')

            else:
                logging.error(f'b1host {name} not found.')
                status = False
        else:
            logging.error(f'service {service} not supported.')
            status = False
        
        return status
    

    def service_process_control(self, service_name="", action=""): 
        '''
        Start or stop a service

        Parameters:
            service_name (str): Name of service
        
        Returns:
            bool 
        '''
        status = False
        actions = [ 'start', 'stop' ]
        service_status = {}
        id = ''
        pool_id = ''
        service_type = ''
        current_state = ''
        desired_state = ''
        
        if action.lower() in actions:
            # Get required service elements
            service_status = self.b1host_service_status(service_name=service_name)
            if service_status:
                currrent_state = service_status.get('composite_status')
                desired_state = service_status.get('desired_status')
                if desired_state == action.lower():
                    logging.info(f'Service {service_name} already set to {desired_state}')
                    logging.info(f'Current state: {current_state}')
                    status = True
                else:
                    id = service_status.get('id')
                    service_type = service_status.get('service_type')
                    logging.debug(f'Service id = {id}')
                    logging.debug(f'service status: {service_name} service is {current_state}')
                    # Check whether service is disabled
                if 'disabled' not in service_status:
                    # Build body
                    body = { 'display_name': service_name, 
                            'servicelications': [ { 'servicelication_type': service_type,
                                                'disabled': '0', 
                                                'state': 
                                                    { 'desired_state': actions[action] } 
                                            } ] } 
                    
                    # Update desired b1host
                    response = self.update('/hosts',
                                        id=id,
                                        body=json.dumps(body))
                    if response.status_code in self.return_codes_ok:
                        logging.debug(f'b1host: {service_name}, service: {service}, service_type: {service_type}')
                        status = True
                    else:
                        logging.error(f'{response.status_code}: {response.text}')
                        status = False
                else:
                    logging.info(f'service: {service} on b1host: {name} {service_status}')
                    status = False

            else:
                logging.error(f'service {service} not supported.')
                status = False
        else:
            logging.error(f'Action: {action} not supported')
            status = False

        return status
    
    """


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 Module to provide class hierachy to simplify access to the BloxOne APIs

 Date Last Updated: 20210215

 Todo:

 Copyright (c) 2021 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
import bloxone
import logging
import requests
import json

# ** Global Vars **
__version__ = '0.1.0'
__author__ = 'Chris Marrison'
__email__ = 'chris@infoblox.com'
__doc__ = 'https://python-bloxone.readthedocs.io/en/latest/'
__license__ = 'BSD'


class b1anycast(bloxone.b1):
    '''
    Class to simplify access to the BloxOne Platform APIs
    '''

    def get(self, objpath, id="", action="", **params):
        '''
        Generic get object wrapper for platform calls

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Optional Object ID
            action (str):   Optional object action, e.g. "nextavailableip"

        Returns:
            response object: Requests response object
        '''

        # Build url
        url = self.anycast_url + objpath
        url = self._use_obj_id(url,id=id)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response

        
    def create(self, objpath, body=""):
        '''
        Generic create object wrapper for platform objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.anycast_url + objpath
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apipost(url, body)

        return response


    def delete(self, objpath, id=""):
        '''
        Generic delete object wrapper for platform objects

        Parameters:
            objpath (str):  Swagger object path
            id (str):       Object id to delete

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.anycast_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apidelete(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for ddi objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.anycast_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response
        
    # *** Helper Methods ***

    def get_id(self, objpath, *, key="", value="", include_path=False):
        '''
        Get object id using key/value pair

        Parameters:
            objpath (str):  Swagger object path
            key (str):      name of key to match
            value (str):    value to match
            include_path (bool): Include path to object id

        Returns:
            id (str):   object id or ""
        '''

        # Local Variables
        id = ""
        filter = key+'=="'+value+'"'
        fields = key + ',id'

        # Make API Call
        response = self.get(objpath, _filter=filter, _fields=fields)

        # Process response
        if response.status_code in self.return_codes_ok:
            obj = response.json()
            # Look for results
            if "results" in obj.keys():
                obj = obj['results']
                if obj:
                    id = obj[0]['id']
                    if not include_path:
                        id = id.rsplit('/',1)[1]
                else:
                    logging.debug("Key {} with value {} not found."
                                  .format(key,value))
            else:
                id = ""
                logging.debug("No results found.")
        else:
            id=""
            logging.debug("HTTP Error occured. {}".format(response.status_code))

        logging.debug("id: {}".format(id)) 

        return id


#!/usr/local/bin/python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
'''
------------------------------------------------------------------------

 Description:

 BloxOne Threat Defence API Wrapper Class and Helpers

 Author: Chris Marrison

 Date Last Updated: 20240701

 Todo:

 Copyright (c) 2020-2021 Chris Marrison / Infoblox

 Redistribution and use in source and binary forms,
 with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

------------------------------------------------------------------------
'''
__version__ = '0.0.1'
__author__ = 'Chris Marrison'
__author_email__ = 'chris@infoblox.com'

import bloxone
import logging
import datetime
import json


# b1td class
class b1insights(bloxone.b1):
    '''
    BloxOne ThreatDefence API Wrapper
    Covers TIDE and Dossier
    '''

    # Generic Methods
    def get(self, objpath, **params):
        '''
        Generic get object wrapper for TIDE data objects

        Parameters:
            objpath (str):  Swagger object path
            action (str):   Optional object action
        
        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.insights_url + objpath
        url = self._use_obj_id(url, id=id)
        url = self._add_params(url, **params)
        logging.debug("URL: {}".format(url))

        response = self._apiget(url)

        return response


    def update(self, objpath, id="", body=""):
        '''
        Generic create object wrapper for insights objects

        Parameters:
            objpath (str):  Swagger object path
            body (str):     JSON formatted data payload

        Returns:
            response object: Requests response object
        '''
        # Build url
        url = self.insights_url + objpath
        url = self._use_obj_id(url, id=id)
        logging.debug("URL: {}".format(url))

        # Make API Call
        response = self._apiput(url, body)

        return response


   # *** Helper Methods ***

    def actor_insights(self, id=""):
        '''
        '''
        return self.get('/actor-insights', id=id)
        
    
    def config_insights(self, id=""):
        '''
        '''
        return self.get('/config-insights/analytics', id=id)


    def policy_check(self):
        '''
        '''
        return self.get('/config-insights/policy-check')
    

    def insights(self, id="", action=""):
        '''
        '''
        return self.get('/insights', id=id, action=action)


    def update_status(self, body=""):
        return self.update('/insights/status', body="")